"""
This module contains the abstract class to implement process metrics.
"""

from datetime import datetime
from pydriller import RepositoryMining

class ProcessMetric:
    """
    Abstract class to implement process metrics
    """

    def __init__(self, path_to_repo: str,
                 since: datetime = None,
                 to: datetime = None,
                 from_commit: str = None,
                 to_commit: str = None):
        """
        :path_to_repo: path to a single repo

        :param datetime since: starting date

        :param datetime to: ending date

        :param str from_commit: starting commit (only if `since` is None)

        :param str to_commit: ending commit (only if `to` is None)
        """

        if not since and not from_commit:
            raise TypeError('You must pass one between since and from_commit')

        if not to and not to_commit:
            raise TypeError('You must pass one between to and to_commit')

        if from_commit and to_commit and from_commit == to_commit:  # Use 'single' param to avoid Warning
            self.repo_miner = RepositoryMining(path_to_repo, single=from_commit)

        else:
            self.repo_miner = RepositoryMining(path_to_repo=path_to_repo,
                                               since=since,
                                               to=to,
                                               from_commit=from_commit,
                                               to_commit=to_commit,
                                               order='reverse')

    def count(self):
        """
        Implement the main functionality of the metric
        """
        return 0



# Copyright 2018 Davide Spadini and Arie van Deursen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pytest

from pydriller.git_repository import GitRepository
from pydriller.domain.commit import Commit, DMMProperty


logging.basicConfig(
    format='%(asctime)s - %(levelname)s - %(message)s', level=logging.INFO)


@pytest.fixture()
def repo():
    path = "test-repos/dmm-test-repo"
    gr = GitRepository(path)
    yield gr
    gr.clear()


# List of (commit_message, dmm_value) pairs
#
# We use unit size to exercise all the various DMM cases
UNIT_SIZE_TEST_DATA = [
    # delta high > 0, delta low = 0 -- always DMM 0.0
    ('Commit with one large method', 0.0),

    # delta high > 0, delta low > 0 -- DMM = ratio
    ('Make large larger, add small method', 0.8),

    # delta high > 0, delta low < 0 --- always DMM 0.0
    ('Make large larger, make small smaller', 0.0),

    # delta high = 0, delta low = 0 --- always DMM 1.0
    ('Modify every line in large method', 1.0),

    # delta high = 0, delta low > 0 --- always DMM 1.0
    ('Make small method a bit larger', 1.0),

    # delta high = 0, delta low < 0 --- alwyas DMM 1.0 (corner case)
    ('Make small smaller', 1.0),

    # delta high < 0, delta low < 0 --- DMM = ratio
    ('Make large smaller, make small smaller', 2/3),

    # delta  high < 0, delta low = 0 -- always 1.0
    ('Make large smaller', 1.0),

    # delta high < 0, delta low > 0 -- always DMM 1.0
    ('Make large smaller, make small larger', 1.0),

    # File 1: large larger; File 2: small larger -- dmm fraction
    ('Increase in one, decrease in other file', 3/4),

    # Method with unit size exactly on the border
    ('Add method with unit size on-point', 1.0),

    # Method with unit size at off point
    ('Increase unit size to risky', 0.0)
]

UNIT_COMPLEXITY_TEST_DATA = [
    # Large method, but no conditional logic
    ('Commit with one large method', 1.0),

    # Method with cyclomatic complexity exactly on the border
    ('Add method with complexity on-point', 1.0),

    # Method with cyclomatic complexity at off point
    ('Increase complexity to risky', 0.0)
]

UNIT_INTERFACING_TEST_DATA = [
    # Large method, but no parameters
    ('Commit with one large method', 1.0),

    # Method with nr of paramters exactly on the border
    ('Add method with interfacing on-point', 1.0),

    # Method with nr of parameters at off point
    ('Increase interfacing to risky', 0.0)
]


def commit_by_msg(repo: GitRepository, msg: str) -> Commit:
    for commit in repo.get_list_commits():
        if commit.msg == msg:
            return commit
    logging.warning('cannot find commit with msg "%s"', msg)
    return None

@pytest.mark.parametrize('msg,dmm', UNIT_SIZE_TEST_DATA)
def test_dmm_unit_size(repo: GitRepository, msg: str, dmm: float):
    commit = commit_by_msg(repo, msg)
    assert commit.dmm_unit_size == dmm

@pytest.mark.parametrize('msg,dmm', UNIT_COMPLEXITY_TEST_DATA)
def test_dmm_unit_complexity(repo: GitRepository, msg: str, dmm: float):
    commit = commit_by_msg(repo, msg)
    assert commit.dmm_unit_complexity == dmm

@pytest.mark.parametrize('msg,dmm', UNIT_INTERFACING_TEST_DATA)
def test_dmm_unit_interfacing(repo: GitRepository, msg: str, dmm: float):
    commit = commit_by_msg(repo, msg)
    assert commit.dmm_unit_interfacing == dmm

def test_delta_profile_modification(repo: GitRepository):
    commit = commit_by_msg(repo, 'Increase unit size to risky')
    mod = commit.modifications[0]
    assert mod._delta_risk_profile(DMMProperty.UNIT_SIZE) == (-15, 16)

def test_delta_profile_commit(repo: GitRepository):
    commit = commit_by_msg(repo, 'Increase in one, decrease in other file')

    m0 = commit.modifications[0]
    assert m0._delta_risk_profile(DMMProperty.UNIT_SIZE) == (0, 1)
    m1 = commit.modifications[1]
    assert m1._delta_risk_profile(DMMProperty.UNIT_SIZE) == (3, 0)

    assert commit._delta_risk_profile(DMMProperty.UNIT_SIZE) == (3, 1)

@pytest.mark.parametrize(
    'dlo,dhi,prop', [
    ( 0,  0, 1.0),
    ( 1,  0, 1.0),
    (-1,  0, 0.0),
    ( 0,  1, 0.0),
    ( 0, -1, 1.0),
    ( 1,  1, 0.5),
    (-1, -1, 0.5),
    ( 1, -1, 1.0),
    (-1,  1, 0.0)
    ])
def test_good_proportion(dlo: int, dhi: int, prop: float):
    assert Commit._good_change_proportion(dlo, dhi) == prop



import pytest
from datetime import datetime
from pydriller.metrics.process.change_set import ChangeSet

TEST_DATA = [
    ('test-repos/pydriller', 'ab36bf45859a210b0eae14e17683f31d19eea041', '71e053f61fc5d31b3e31eccd9c79df27c31279bf', 13, 8)
]

@pytest.mark.parametrize('path_to_repo, from_commit, to_commit, expected_max, expected_avg', TEST_DATA)
def test_with_commits(path_to_repo, from_commit, to_commit, expected_max, expected_avg):
    metric = ChangeSet(path_to_repo=path_to_repo,
                       from_commit=from_commit,
                       to_commit=to_commit)

    actual_max = metric.max()
    actual_avg = metric.avg()

    assert actual_max == expected_max
    assert actual_avg == expected_avg


TEST_DATA = [
    ('test-repos/pydriller', datetime(2018, 3, 21), datetime(2018, 3, 27), 13, 8),
    ('test-repos/pydriller', datetime(2018, 3, 23), datetime(2018, 3, 23), 0, 0)
]

@pytest.mark.parametrize('path_to_repo, since, to, expected_max, expected_avg', TEST_DATA)
def test_with_dates(path_to_repo, since, to, expected_max, expected_avg):
    metric = ChangeSet(path_to_repo=path_to_repo,
                       since=since,
                       to=to)

    actual_max = metric.max()
    actual_avg = metric.avg()

    assert actual_max == expected_max
    assert actual_avg == expected_avg



from pathlib import Path
from datetime import datetime

import pytest

from pydriller.metrics.process.contributors_count import ContributorsCount

TEST_DATA = [
   ('test-repos/pydriller', 'pydriller/git_repository.py', 'e9854bbea1cb7b7f06cbb559f7b06724d11ae1e5', 'e9854bbea1cb7b7f06cbb559f7b06724d11ae1e5', 0),
   ('test-repos/pydriller', 'pydriller/git_repository.py', 'e9854bbea1cb7b7f06cbb559f7b06724d11ae1e5', 'ab36bf45859a210b0eae14e17683f31d19eea041', 1),
   ('test-repos/pydriller', 'pydriller/git_repository.py', '4af3839eb5ea5969f42142529a7a5526739fa570', 'ab36bf45859a210b0eae14e17683f31d19eea041', 2)
]

@pytest.mark.parametrize('path_to_repo, filepath, from_commit, to_commit, expected', TEST_DATA)
def test_with_commits(path_to_repo, filepath, from_commit, to_commit, expected):
    metric = ContributorsCount(path_to_repo=path_to_repo,
                               from_commit=from_commit,
                               to_commit=to_commit)

    count = metric.count_minor()
    filepath = str(Path(filepath))
    assert count[filepath] == expected


TEST_DATA = [
   ('test-repos/pydriller', 'pydriller/git_repository.py', datetime(2018, 8, 1), datetime(2018, 8, 2), 0),
   ('test-repos/pydriller', 'pydriller/git_repository.py', datetime(2018, 3, 21), datetime(2018, 8, 2), 1),
   ('test-repos/pydriller', 'pydriller/git_repository.py', datetime(2018, 3, 21), datetime(2019, 1, 14, 10), 2)
]

@pytest.mark.parametrize('path_to_repo, filepath, since, to, expected', TEST_DATA)
def test_with_dates(path_to_repo, filepath, since, to, expected):
    metric = ContributorsCount(path_to_repo=path_to_repo,
                               since=since,
                               to=to)

    count = metric.count_minor()
    filepath = str(Path(filepath))
    assert count[filepath] == expected



"""
Module that calculates the number of commits made to a file.
"""

from pydriller import ModificationType
from pydriller.metrics.process.process_metric import ProcessMetric

class CommitsCount(ProcessMetric):
    """
    This class is responsible to implement the Commit Count metric to \
    measure the number of commits made to a file
    """

    def count(self):
        files = {}
        renamed_files = {}  # To keep track of renamed files

        for commit in self.repo_miner.traverse_commits():

            for modified_file in commit.modifications:

                filepath = renamed_files.get(modified_file.new_path,
                                             modified_file.new_path)
                if modified_file.change_type == ModificationType.RENAME:
                    renamed_files[modified_file.old_path] = filepath

                files[filepath] = files.get(filepath, 0) + 1

        return files



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import platform
import sys
import psutil
from pydriller import RepositoryMining
from datetime import datetime

logging.basicConfig(level=logging.WARNING)
skip_remote = True


def test_memory(caplog):
    if skip_remote:
        return
    caplog.set_level(logging.WARNING)

    logging.warning("Starting with nothing...")
    diff_with_nothing, all_commits_with_nothing = mine(0)

    logging.warning("Starting with everything...")
    diff_with_everything, all_commits_with_everything = mine(1)

    logging.warning("Starting with metrics...")
    diff_with_metrics, all_commits_with_metrics = mine(2)

    max_values = [max(all_commits_with_nothing),
                  max(all_commits_with_everything),
                  max(all_commits_with_metrics)]
    logging.warning("Max values are: {}".format(max_values))

    minutes_with_everything = (diff_with_everything.seconds % 3600) // 60
    minutes_with_metrics = (diff_with_metrics.seconds % 3600) // 60

    logging.warning("TIME: With nothing: {}:{}:{} ({} commits/sec), "
                    "with everything: {}:{}:{}  ({} commits/sec), "
                    "with metrics: {}:{}:{}  ({} commits/sec)".format(
        diff_with_nothing.seconds // 3600,
        (diff_with_nothing.seconds % 3600) // 60,
        diff_with_nothing.seconds % 60,
        973 // diff_with_nothing.seconds if diff_with_nothing.seconds != 0 else 0,
        diff_with_everything.seconds // 3600,
        (diff_with_everything.seconds % 3600) // 60,
        diff_with_everything.seconds % 60,
        973 // diff_with_everything.seconds,
        diff_with_metrics.seconds // 3600,
        (diff_with_metrics.seconds % 3600) // 60,
        diff_with_metrics.seconds % 60,
        973 // diff_with_metrics.seconds
    ))

    if any(val > 250 for val in max_values) or \
            minutes_with_everything >= 1 or \
            minutes_with_metrics >= 7:
        # if to analyze 1000 commits requires more than 250MB of RAM,
        # more than 1 minute without metrics or
        # 7 minutes with metrics, print it
        log(diff_with_nothing, all_commits_with_nothing,
            diff_with_everything, all_commits_with_everything,
            diff_with_metrics, all_commits_with_metrics)

    assert 973 == len(all_commits_with_nothing) == len(all_commits_with_everything) == len(all_commits_with_metrics)


def log(diff_with_nothing, all_commits_with_nothing,
        diff_with_everything, all_commits_with_everything,
        diff_with_metrics, all_commits_with_metrics):
    text = "*PYTHON V{}.{} - System: {}*\n" \
           "*Max memory (MB)*\n" \
           "With nothing: {}, with everything: {}, with metrics: {}\n" \
           "*Min memory (MB)*\n" \
           "With nothing: {}, with everything: {}, with metrics: {} \n" \
           "*Time*\n" \
           "With nothing: {}:{}:{}, with everything: {}:{}:{}, with metrics: {}:{}:{} \n" \
           "*Total number of commits*: {}\n" \
           "*Commits per second:*\n" \
           "With nothing: {}, with everything: {}, with metrics: {}"

    print(text.format(
        sys.version_info[0], sys.version_info[1], platform.system(),
        max(all_commits_with_nothing), max(all_commits_with_everything), max(all_commits_with_metrics),
        min(all_commits_with_nothing), min(all_commits_with_everything), min(all_commits_with_metrics),
        diff_with_nothing.seconds // 3600, (diff_with_nothing.seconds % 3600) // 60, diff_with_nothing.seconds % 60,
        diff_with_everything.seconds // 3600, (diff_with_everything.seconds % 3600) // 60,
        diff_with_everything.seconds % 60,
        diff_with_metrics.seconds // 3600, (diff_with_metrics.seconds % 3600) // 60, diff_with_metrics.seconds % 60,
        len(all_commits_with_nothing),
        len(all_commits_with_nothing) / diff_with_nothing.seconds,
        len(all_commits_with_everything) / diff_with_everything.seconds,
        len(all_commits_with_metrics) / diff_with_metrics.seconds
    ))


def mine(_type):
    p = psutil.Process(os.getpid())
    dt1 = datetime(2017, 1, 1)
    dt2 = datetime(2017, 7, 1)
    all_commits = []

    start = datetime.now()
    for commit in RepositoryMining('test-repos-hadoop/hadoop',
                                   since=dt1,
                                   to=dt2).traverse_commits():
        memory = p.memory_info()[0] / (2 ** 20)
        all_commits.append(memory)

        h = commit.author.name

        if _type == 0:
            continue

        for mod in commit.modifications:
            dd = mod.diff

            if _type == 1:
                continue

            if mod.filename.endswith('.java'):
                cc = mod.complexity

    end = datetime.now()

    diff = end - start

    return diff, all_commits



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains all the classes regarding a specific commit, such as
Commit, Modification,
ModificationType and Method.
"""
import logging
from _datetime import datetime
from enum import Enum
from pathlib import Path
from typing import List, Set, Dict, Tuple

import lizard
from git import Diff, Git, Commit as GitCommit, NULL_TREE

from pydriller.domain.developer import Developer

logger = logging.getLogger(__name__)


class ModificationType(Enum):
    """
    Type of Modification. Can be ADD, COPY, RENAME, DELETE, MODIFY or UNKNOWN.
    """

    ADD = 1
    COPY = 2
    RENAME = 3
    DELETE = 4
    MODIFY = 5
    UNKNOWN = 6


class DMMProperty(Enum):
    """
    Maintainability properties of the Delta Maintainability Model.
    """
    UNIT_SIZE = 1
    UNIT_COMPLEXITY = 2
    UNIT_INTERFACING = 3


class Method:  # pylint: disable=R0902
    """
    This class represents a method in a class. Contains various information
    extracted through Lizard.
    """

    def __init__(self, func):
        """
        Initialize a method object. This is calculated using Lizard: it parses
        the source code of all the modifications in a commit, extracting
        information of the methods contained in the file (if the file is a
        source code written in one of the supported programming languages).
        """

        self.name = func.name
        self.long_name = func.long_name
        self.filename = func.filename
        self.nloc = func.nloc
        self.complexity = func.cyclomatic_complexity
        self.token_count = func.token_count
        self.parameters = func.parameters
        self.start_line = func.start_line
        self.end_line = func.end_line
        self.fan_in = func.fan_in
        self.fan_out = func.fan_out
        self.general_fan_out = func.general_fan_out
        self.length = func.length
        self.top_nesting_level = func.top_nesting_level

    def __eq__(self, other):
        return self.name == other.name and self.parameters == other.parameters

    def __hash__(self):
        # parameters are used in hashing in order to
        # prevent collisions when overloading method names
        return hash(('name', self.name,
                     'long_name', self.long_name,
                     'params', (x for x in self.parameters)))

    UNIT_SIZE_LOW_RISK_THRESHOLD = 15
    """
    Threshold used in the Delta Maintainability Model to establish whether a method
    is low risk in terms of its size.
    The procedure to obtain the threshold is described in the
    :ref:`PyDriller documentation <Properties>`.
    """

    UNIT_COMPLEXITY_LOW_RISK_THRESHOLD = 5
    """
    Threshold used in the Delta Maintainability Model to establish whether a method
    is low risk in terms of its cyclomatic complexity.
    The procedure to obtain the threshold is described in the
    :ref:`PyDriller documentation <Properties>`.
    """

    UNIT_INTERFACING_LOW_RISK_THRESHOLD = 2
    """
    Threshold used in the Delta Maintainability Model to establish whether a method
    is low risk in terms of its interface.
    The procedure to obtain the threshold is described in the
    :ref:`PyDriller documentation <Properties>`.
    """

    def is_low_risk(self, dmm_prop: DMMProperty) -> bool:
        """
        Predicate indicating whether this method is low risk in terms of
        the given property.

        :param dmm_prop: Property according to which this method is considered risky.
        :return: True if and only if the method is considered low-risk w.r.t. this property.
        """
        if dmm_prop is DMMProperty.UNIT_SIZE:
            return self.nloc <= Method.UNIT_SIZE_LOW_RISK_THRESHOLD
        if dmm_prop is DMMProperty.UNIT_COMPLEXITY:
            return self.complexity <= Method.UNIT_COMPLEXITY_LOW_RISK_THRESHOLD
        assert dmm_prop is DMMProperty.UNIT_INTERFACING
        return len(self.parameters) <= Method.UNIT_INTERFACING_LOW_RISK_THRESHOLD


class Modification:  # pylint: disable=R0902
    """
    This class contains information regarding a modified file in a commit.
    """

    def __init__(self, old_path: str, new_path: str,
                 change_type: ModificationType,
                 diff_and_sc: Dict[str, str]):
        """
        Initialize a modification. A modification carries on information
        regarding the changed file. Normally, you shouldn't initialize a new
        one.
        """
        self._old_path = Path(old_path) if old_path is not None else None
        self._new_path = Path(new_path) if new_path is not None else None
        self.change_type = change_type
        self.diff = diff_and_sc['diff']
        self.source_code = diff_and_sc['source_code']
        self.source_code_before = diff_and_sc['source_code_before']

        self._nloc = None
        self._complexity = None
        self._token_count = None
        self._function_list = []
        self._function_list_before = []

    @property
    def added(self) -> int:
        """
        Return the total number of added lines in the file.

        :return: int lines_added
        """
        added = 0
        for line in self.diff.replace('\r', '').split("\n"):
            if line.startswith('+') and not line.startswith('+++'):
                added += 1
        return added

    @property
    def removed(self):
        """
        Return the total number of deleted lines in the file.

        :return: int lines_deleted
        """
        removed = 0
        for line in self.diff.replace('\r', '').split("\n"):
            if line.startswith('-') and not line.startswith('---'):
                removed += 1
        return removed

    @property
    def old_path(self):
        """
        Old path of the file. Can be None if the file is added.

        :return: str old_path
        """
        if self._old_path is not None:
            return str(self._old_path)
        return None

    @property
    def new_path(self):
        """
        New path of the file. Can be None if the file is deleted.

        :return: str new_path
        """
        if self._new_path is not None:
            return str(self._new_path)
        return None

    @property
    def filename(self) -> str:
        """
        Return the filename. Given a path-like-string (e.g.
        "/Users/dspadini/pydriller/myfile.py") returns only the filename
        (e.g. "myfile.py")

        :return: str filename
        """
        if self._new_path is not None and str(self._new_path) != "/dev/null":
            path = self._new_path
        else:
            path = self._old_path

        return path.name

    @property
    def nloc(self) -> int:
        """
        Calculate the LOC of the file.

        :return: LOC of the file
        """
        self._calculate_metrics()
        return self._nloc

    @property
    def complexity(self) -> int:
        """
        Calculate the Cyclomatic Complexity of the file.

        :return: Cyclomatic Complexity of the file
        """
        self._calculate_metrics()
        return self._complexity

    @property
    def token_count(self) -> int:
        """
        Calculate the token count of functions.

        :return: token count
        """
        self._calculate_metrics()
        return self._token_count

    @property
    def diff_parsed(self) -> Dict[str, List[Tuple[int, str]]]:
        """
        Returns a dictionary with the added and deleted lines.
        The dictionary has 2 keys: "added" and "deleted", each containing the
        corresponding added or deleted lines. For both keys, the value is a
        list of Tuple (int, str), corresponding to (number of line in the file,
        actual line).

        :return: Dictionary
        """
        lines = self.diff.split('\n')
        modified_lines = {'added': [], 'deleted': []}

        count_deletions = 0
        count_additions = 0

        for line in lines:
            line = line.rstrip()
            count_deletions += 1
            count_additions += 1

            if line.startswith('@@'):
                count_deletions, count_additions = self._get_line_numbers(line)

            if line.startswith('-'):
                modified_lines['deleted'].append((count_deletions, line[1:]))
                count_additions -= 1

            if line.startswith('+'):
                modified_lines['added'].append((count_additions, line[1:]))
                count_deletions -= 1

            if line == r'\ No newline at end of file':
                count_deletions -= 1
                count_additions -= 1

        return modified_lines

    @staticmethod
    def _get_line_numbers(line):
        token = line.split(" ")
        numbers_old_file = token[1]
        numbers_new_file = token[2]
        delete_line_number = int(numbers_old_file.split(",")[0]
                                 .replace("-", "")) - 1
        additions_line_number = int(numbers_new_file.split(",")[0]) - 1
        return delete_line_number, additions_line_number

    @property
    def methods(self) -> List[Method]:
        """
        Return the list of methods in the file. Every method
        contains various information like complexity, loc, name,
        number of parameters, etc.

        :return: list of methods
        """
        self._calculate_metrics()
        return self._function_list

    @property
    def methods_before(self) -> List[Method]:
        """
        Return the list of methods in the file before the
        change happened. Each method will have all specific
        info, e.g. complexity, loc, name, etc.

        :return: list of methods
        """
        self._calculate_metrics(include_before=True)
        return self._function_list_before

    @property
    def changed_methods(self) -> List[Method]:
        """
        Return the list of methods that were changed. This analysis
        is more complex because Lizard runs twice: for methods before
        and after the change

        :return: list of methods
        """
        new_methods = self.methods
        old_methods = self.methods_before
        added = self.diff_parsed['added']
        deleted = self.diff_parsed['deleted']

        methods_changed_new = {y for x in added for y in new_methods if
                               y.start_line <= x[0] <= y.end_line}
        methods_changed_old = {y for x in deleted for y in old_methods if
                               y.start_line <= x[0] <= y.end_line}

        return list(methods_changed_new.union(methods_changed_old))

    @staticmethod
    def _risk_profile(methods: List[Method], dmm_prop: DMMProperty) -> Tuple[int, int]:
        """
        Return the risk profile of the set of methods, with two bins: risky, or non risky.
        The risk profile is a pair (v_low, v_high), where
        v_low is the volume of the low risk methods in the list, and
        v_high is the volume of the high risk methods in the list.

        :param methods: List of methods for which risk profile is to be determined
        :param dmm_prop: Property indicating the type of risk
        :return: total risk profile for methods according to property.
        """
        low = sum([m.nloc for m in methods if m.is_low_risk(dmm_prop)])
        high = sum([m.nloc for m in methods if not m.is_low_risk(dmm_prop)])
        return low, high

    def _delta_risk_profile(self, dmm_prop: DMMProperty) -> Tuple[int, int]:
        """
        Return the delta risk profile of this commit, which a pair (dv1, dv2), where
        dv1 is the total change in volume (lines of code) of low risk methods, and
        dv2 is the total change in volume of the high risk methods.

        :param dmm_prop: Property indicating the type of risk
        :return: total delta risk profile for this property.
        """
        low_before, high_before = self._risk_profile(self.methods_before, dmm_prop)
        low_after, high_after = self._risk_profile(self.methods, dmm_prop)
        return low_after - low_before, high_after - high_before

    def _calculate_metrics(self, include_before=False):
        """
        :param include_before: either to compute the metrics
        for source_code_before, i.e. before the change happened
        """
        if self.source_code and self._nloc is None:
            analysis = lizard.analyze_file.analyze_source_code(
                self.filename, self.source_code)
            self._nloc = analysis.nloc
            self._complexity = analysis.CCN
            self._token_count = analysis.token_count

            for func in analysis.function_list:
                self._function_list.append(Method(func))

        if include_before and self.source_code_before and \
                not self._function_list_before:
            anal = lizard.analyze_file.analyze_source_code(
                self.filename, self.source_code_before)

            self._function_list_before = [
                Method(x) for x in anal.function_list]

    def __eq__(self, other):
        if not isinstance(other, Modification):
            return NotImplemented
        if self is other:
            return True
        return self.__dict__ == other.__dict__


class Commit:
    """
    Class representing a Commit. Contains all the important information such
    as hash, author, dates, and modified files.
    """

    def __init__(self, commit: GitCommit, conf) -> None:
        """
        Create a commit object.

        :param commit: GitPython Commit object
        :param conf: Configuration class
        """
        self._c_object = commit

        self._modifications = None
        self._branches = None
        self._conf = conf

    @property
    def hash(self) -> str:
        """
        Return the SHA of the commit.

        :return: str hash
        """
        return self._c_object.hexsha

    @property
    def author(self) -> Developer:
        """
        Return the author of the commit as a Developer object.

        :return: author
        """
        return Developer(self._c_object.author.name,
                         self._c_object.author.email)

    @property
    def committer(self) -> Developer:
        """
        Return the committer of the commit as a Developer object.

        :return: committer
        """
        return Developer(self._c_object.committer.name,
                         self._c_object.committer.email)

    @property
    def project_name(self) -> str:
        """
        Return the project name.

        :return: project name
        """
        return Path(self._conf.get('path_to_repo')).name

    @property
    def author_date(self) -> datetime:
        """
        Return the authored datetime.

        :return: datetime author_datetime
        """
        return self._c_object.authored_datetime

    @property
    def committer_date(self) -> datetime:
        """
        Return the committed datetime.

        :return: datetime committer_datetime
        """
        return self._c_object.committed_datetime

    @property
    def author_timezone(self) -> int:
        """
        Author timezone expressed in seconds from epoch.

        :return: int timezone
        """
        return self._c_object.author_tz_offset

    @property
    def committer_timezone(self) -> int:
        """
        Author timezone expressed in seconds from epoch.

        :return: int timezone
        """
        return self._c_object.committer_tz_offset

    @property
    def msg(self) -> str:
        """
        Return commit message.

        :return: str commit_message
        """
        return self._c_object.message.strip()

    @property
    def parents(self) -> List[str]:
        """
        Return the list of parents SHAs.

        :return: List[str] parents
        """
        parents = []
        for p in self._c_object.parents:
            parents.append(p.hexsha)
        return parents

    @property
    def merge(self) -> bool:
        """
        Return True if the commit is a merge, False otherwise.

        :return: bool merge
        """
        return len(self._c_object.parents) > 1

    @property
    def modifications(self) -> List[Modification]:
        """
        Return a list of modified files.

        :return: List[Modification] modifications
        """
        if self._modifications is None:
            self._modifications = self._get_modifications()

        return self._modifications

    def _get_modifications(self):
        options = {}
        if self._conf.get('histogram'):
            options['histogram'] = True

        if self._conf.get('skip_whitespaces'):
            options['w'] = True

        if len(self.parents) == 1:
            # the commit has a parent
            diff_index = self._c_object.parents[0].diff(self._c_object,
                                                        create_patch=True,
                                                        **options)
        elif len(self.parents) > 1:
            # if it's a merge commit, the modified files of the commit are the
            # conflicts. This because if the file is not in conflict,
            # pydriller will visit the modification in one of the previous
            # commits. However, parsing the output of a combined diff (that
            # returns the list of conflicts) is challenging: so, right now,
            # I will return an empty array, in the meanwhile I will try to
            # find a way to parse the output.
            # c_git = Git(str(self.project_path))
            # d = c_git.diff_tree("--cc", commit.hexsha, '-r', '--abbrev=40',
            #                     '--full-index', '-M', '-p', '--no-color')
            diff_index = []
        else:
            # this is the first commit of the repo. Comparing it with git
            # NULL TREE
            diff_index = self._c_object.diff(NULL_TREE,
                                             create_patch=True,
                                             **options)

        return self._parse_diff(diff_index)

    def _parse_diff(self, diff_index) -> List[Modification]:
        modifications_list = []
        for diff in diff_index:
            old_path = diff.a_path
            new_path = diff.b_path
            change_type = self._from_change_to_modification_type(diff)

            diff_and_sc = {
                'diff': self._get_decoded_str(diff.diff),
                'source_code_before': self._get_decoded_sc_str(
                    diff.a_blob),
                'source_code': self._get_decoded_sc_str(
                    diff.b_blob)
            }

            modifications_list.append(Modification(old_path, new_path,
                                                   change_type, diff_and_sc))

        return modifications_list

    def _get_decoded_str(self, diff):
        try:
            return diff.decode('utf-8', 'ignore')
        except (UnicodeDecodeError, AttributeError, ValueError):
            logger.debug('Could not load the diff of a '
                         'file in commit %s', self._c_object.hexsha)
            return None

    def _get_decoded_sc_str(self, diff):
        try:
            return diff.data_stream.read().decode('utf-8', 'ignore')
        except (UnicodeDecodeError, AttributeError, ValueError):
            logger.debug('Could not load source code of a '
                         'file in commit %s', self._c_object.hexsha)
            return None

    @property
    def in_main_branch(self) -> bool:
        """
        Return True if the commit is in the main branch, False otherwise.

        :return: bool in_main_branch
        """
        return self._conf.get('main_branch') in self.branches

    @property
    def branches(self) -> Set[str]:
        """
        Return the set of branches that contain the commit.

        :return: set(str) branches
        """
        if self._branches is None:
            self._branches = self._get_branches()

        return self._branches

    def _get_branches(self):
        c_git = Git(str(self._conf.get('path_to_repo')))
        branches = set()
        for branch in set(c_git.branch('--contains', self.hash).split('\n')):
            branches.add(branch.strip().replace('* ', ''))
        return branches

    @property
    def dmm_unit_size(self) -> float:
        """
        Return the Delta Maintainability Model (DMM) metric value for the unit size property.

        It represents the proportion (between 0.0 and 1.0) of maintainability improving
        change, when considering the lengths of the modified methods.

        It rewards (value close to 1.0) modifications to low-risk (small) methods,
        or spliting risky (large) ones.
        It penalizes (value close to 0.0) working on methods that remain large
        or get larger.

        :return: The DMM value (between 0.0 and 1.0) for method size in this commit.
        """
        return self._delta_maintainability(DMMProperty.UNIT_SIZE)

    @property
    def dmm_unit_complexity(self) -> float:
        """
        Return the Delta Maintainability Model (DMM) metric value for the unit complexity property.

        It represents the proportion (between 0.0 and 1.0) of maintainability improving
        change, when considering the cyclomatic complexity of the modified methods.

        It rewards (value close to 1.0) modifications to low-risk (low complexity) methods,
        or spliting risky (highly complex) ones.
        It penalizes (value close to 0.0) working on methods that remain complex
        or get more complex.

        :return: The DMM value (between 0.0 and 1.0) for method complexity in this commit.
        """
        return self._delta_maintainability(DMMProperty.UNIT_COMPLEXITY)

    @property
    def dmm_unit_interfacing(self):
        """
        Return the Delta Maintainability Model (DMM) metric value for the unit interfacing property.

        It represents the proportion (between 0.0 and 1.0) of maintainability improving
        change, when considering the interface (number of parameters) of the modified methods.

        It rewards (value close to 1.0) modifications to low-risk (with  few parameters) methods,
        or spliting risky (with many parameters) ones.
        It penalizes (value close to 0.0) working on methods that continue to have
        or are extended with too many parameters.

        :return: The dmm value (between 0.0 and 1.0) for method interfacing in this commit.
        """
        return self._delta_maintainability(DMMProperty.UNIT_INTERFACING)

    def _delta_maintainability(self, dmm_prop: DMMProperty) -> float:
        """
        Compute the Delta Maintainability Model (DMM) value for the given risk predicate.
        The DMM value is computed as the proportion of good change in the commit:
        Good changes: Adding low risk code or removing high risk codee.
        Bad changes: Adding high risk code or removing low risk code.
        Furthermore, the special case where only low risk code is removed is also given the highest score.

        :param dmm_prop: Property indicating the type of risk
        :return: dmm value (between 0.0 and 1.0) for the property represented in the property.
        """
        (delta_low, delta_high) = self._delta_risk_profile(dmm_prop)
        dmm = self._good_change_proportion(delta_low, delta_high)
        if delta_low < 0 and delta_high == 0:
            assert dmm == 0.0
            # special case where removal of good (low risk) code is OK.
            # re-adjust dmm accordingly:
            dmm = 1.0
        assert 0.0 <= dmm <= 1.0
        return dmm

    def _delta_risk_profile(self, dmm_prop: DMMProperty) -> Tuple[int, int]:
        """
        Return the delta risk profile of this commit, which a pair (dv1, dv2), where
        dv1 is the total change in volume (lines of code) of low risk methods, and
        dv2 is the total change in volume of the high risk methods.

        :param dmm_prop: Property indicating the type of risk
        :return: total delta risk profile for this commit.
        """
        deltas = [mod._delta_risk_profile(dmm_prop) for mod in self.modifications]  #pylint: disable=W0212
        delta_low = sum(d[0] for d in deltas)
        delta_high = sum(d[1] for d in deltas)
        return delta_low, delta_high

    @staticmethod
    def _good_change_proportion(low_risk_delta: int, high_risk_delta: int) -> float:
        """
        Given a delta risk profile, compute the proportion of "good" change in the total change.
        Increasing low risk code, or decreasing high risk code, is considered good.
        Other types of changes are considered not good.

        :return: proportion of good change in total change.
        """
        bad_change, good_change = (0, 0)

        if low_risk_delta >= 0:
            good_change = low_risk_delta
        else:
            bad_change = abs(low_risk_delta)
        if high_risk_delta >= 0:
            bad_change += high_risk_delta
        else:
            good_change += abs(high_risk_delta)

        assert good_change >= 0 and bad_change >= 0

        total_change = good_change + bad_change
        result = good_change / total_change if total_change > 0 else 1.0
        assert 0.0 <= result <= 1.0

        return result

    # pylint disable=R0902
    @staticmethod
    def _from_change_to_modification_type(diff: Diff):
        if diff.new_file:
            return ModificationType.ADD
        if diff.deleted_file:
            return ModificationType.DELETE
        if diff.renamed_file:
            return ModificationType.RENAME
        if diff.a_blob and diff.b_blob and diff.a_blob != diff.b_blob:
            return ModificationType.MODIFY

        return ModificationType.UNKNOWN

    def __eq__(self, other):
        if not isinstance(other, Commit):
            return NotImplemented
        if self is other:
            return True

        return self.__dict__ == other.__dict__



from pathlib import Path
from datetime import datetime

import pytest
from pydriller.metrics.process.history_complexity import HistoryComplexity

TEST_DATA = [
    ('test-repos/pydriller', 'scm/git_repository.py', '90ca34ebfe69629cb7f186a1582fc38a73cc572e', '90ca34ebfe69629cb7f186a1582fc38a73cc572e', 40.49),
    ('test-repos/pydriller', 'scm/git_repository.py', '71e053f61fc5d31b3e31eccd9c79df27c31279bf', '90ca34ebfe69629cb7f186a1582fc38a73cc572e', 47.05),
    ('https://github.com/geerlingguy/ansible-role-solr', 'tasks/main.yml', '7fb350c30be1124b51aab4a88352428e0a853b9a', '678429591513fe86045e892a1da680c8ac36e00f', .0)
]

@pytest.mark.parametrize('path_to_repo, filepath, from_commit, to_commit, expected', TEST_DATA)
def test_with_commits(path_to_repo, filepath, from_commit, to_commit, expected):
    metric = HistoryComplexity(path_to_repo=path_to_repo,
                               from_commit=from_commit,
                               to_commit=to_commit)
    
    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected


TEST_DATA = [
    ('test-repos/pydriller', 'scm/git_repository.py', datetime(2018, 3, 22, 11, 30), datetime(2018, 3, 23), 40.49),
    ('test-repos/pydriller', 'scm/git_repository.py', datetime(2018, 3, 22, 11, 30), datetime(2018, 3, 27), 47.05),
]

@pytest.mark.parametrize('path_to_repo, filepath, since, to, expected', TEST_DATA)
def test_with_dates(path_to_repo, filepath, since, to, expected):
    metric = HistoryComplexity(path_to_repo=path_to_repo,
                               since=since,
                               to=to)
    
    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from pydriller import GitRepository


@pytest.fixture
def commit():
    return None


@pytest.fixture()
def modification(commit):
    gr = GitRepository("test-repos/diff")
    yield gr.get_commit(commit).modifications[0]
    gr.clear()


@pytest.mark.parametrize('commit', ["9a985d4a12a3a12f009ef39750fd9b2187b766d1"])
def test_extract_line_number_and_content(modification):
    added = modification.diff_parsed['added']
    deleted = modification.diff_parsed['deleted']

    assert (127, '            RevCommit root = rw.parseCommit(headId);') in deleted
    assert (128, '            rw.sort(RevSort.REVERSE);') in deleted
    assert (129, '            rw.markStart(root);') in deleted
    assert (130, '            RevCommit lastCommit = rw.next();') in deleted
    assert (131, '            throw new RuntimeException("Changing this line " + path);') in added


@pytest.mark.parametrize('commit', ["f45ee2f8976d5f018a1e4ec83eb4556a3df8b0a5"])
def test_additions(modification):
    added = modification.diff_parsed['added']
    deleted = modification.diff_parsed['deleted']

    assert (127, '            RevCommit root = rw.parseCommit(headId);') in added
    assert (128, '            rw.sort(RevSort.REVERSE);') in added
    assert (129, '            rw.markStart(root);') in added
    assert (130, '            RevCommit lastCommit = rw.next();') in added
    assert (131, '') in added
    assert len(deleted) == 0
    assert len(added) == 5


@pytest.mark.parametrize('commit', ["147c7ce9f725a0e259d63f0bf4e6c8ac085ff8c8"])
def test_deletions(modification):
    added = modification.diff_parsed['added']
    deleted = modification.diff_parsed['deleted']

    assert (184, '            List<ChangeSet> allCs = new ArrayList<>();') in deleted
    assert (221, '    private GregorianCalendar convertToDate(RevCommit revCommit) {') in deleted
    assert (222, '        GregorianCalendar date = new GregorianCalendar();') in deleted
    assert (223, '        date.setTimeZone(revCommit.getAuthorIdent().getTimeZone());') in deleted
    assert (224, '        date.setTime(revCommit.getAuthorIdent().getWhen());') in deleted
    assert (225, '') in deleted
    assert (226, '        return date;') in deleted
    assert (227, '    }') in deleted
    assert (228, '') in deleted
    assert (301, '        if(!collectConfig.isCollectingBranches())') in deleted
    assert (302, '            return new HashSet<>();') in deleted
    assert (303, '') in deleted
    assert len(deleted) == 12
    assert len(added) == 0


def test_diff_no_newline():
    """
    If a file ends without a newline git represents this with the additional line
        \\ No newline at end of file
    in diffs. This test asserts these additional lines are parsed correctly.
    """
    gr = GitRepository('test-repos/no_newline')

    mod = gr.get_commit('52a78c1ee5d100528eccba0a3d67371dbd22d898').modifications[0]
    added = mod.diff_parsed['added']
    deleted = mod.diff_parsed['deleted']

    assert (1, 'test1') in deleted  # is considered as deleted as a 'newline' command is added
    assert (1, 'test1') in added  # now with added 'newline'
    assert (2, 'test2') in added

    gr.clear()



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module includes 1 class, RepositoryMining, main class of PyDriller.
"""

import logging
import os
import tempfile
from datetime import datetime
from pathlib import Path
from typing import List, Generator, Union

from git import Repo

from pydriller.domain.commit import Commit
from pydriller.git_repository import GitRepository
from pydriller.utils.conf import Conf

logger = logging.getLogger(__name__)


class RepositoryMining:
    """
    This is the main class of PyDriller, responsible for running the study.
    """

    # pylint: disable=R0902,R0913,R0914
    def __init__(self, path_to_repo: Union[str, List[str]],
                 single: str = None,
                 since: datetime = None, to: datetime = None,
                 from_commit: str = None, to_commit: str = None,
                 from_tag: str = None, to_tag: str = None,
                 include_refs: bool = False,
                 include_remotes: bool = False,
                 reversed_order: bool = False,
                 only_in_branch: str = None,
                 only_modifications_with_file_types: List[str] = None,
                 only_no_merge: bool = False,
                 only_authors: List[str] = None,
                 only_commits: List[str] = None,
                 only_releases: bool = False,
                 filepath: str = None,
                 histogram_diff: bool = False,
                 skip_whitespaces: bool = False,
                 clone_repo_to: str = None,
                 order: str = None):
        """
        Init a repository mining. The only required parameter is
        "path_to_repo": to analyze a single repo, pass the absolute path to
        the repo; if you need to analyze more repos, pass a list of absolute
        paths.

        Furthermore, PyDriller supports local and remote repositories: if
        you pass a path to a repo, PyDriller will run the study on that
        repo; if you pass an URL, PyDriller will clone the repo in a
        temporary folder, run the study, and delete the temporary folder.

        :param Union[str,List[str]] path_to_repo: absolute path (or list of
            absolute paths) to the repository(ies) to analyze
        :param str single: hash of a single commit to analyze
        :param datetime since: starting date
        :param datetime to: ending date
        :param str from_commit: starting commit (only if `since` is None)
        :param str to_commit: ending commit (only if `to` is None)
        :param str from_tag: starting the analysis from specified tag (only
            if `since` and `from_commit` are None)
        :param str to_tag: ending the analysis from specified tag (only if
            `to` and `to_commit` are None)
        :param bool include_refs: whether to include refs and HEAD in commit analysis
        :param bool include_remotes: whether to include remote commits in analysis
        :param bool reversed_order: whether the commits should be analyzed
            in reversed order (**DEPRECATED**)
        :param str only_in_branch: only commits in this branch will be analyzed
        :param List[str] only_modifications_with_file_types: only
            modifications with that file types will be analyzed
        :param bool only_no_merge: if True, merges will not be analyzed
        :param List[str] only_authors: only commits of these authors will be
            analyzed (the check is done on the username, NOT the email)
        :param List[str] only_commits: only these commits will be analyzed
        :param str filepath: only commits that modified this file will be
            analyzed
        :param str order: order of commits. It can be one of: 'date-order',
            'author-date-order', 'topo-order', or 'reverse'. Default is reverse.
        """
        if only_modifications_with_file_types is not None:
            only_modifications_with_file_types = set(only_modifications_with_file_types)
        if only_commits is not None:
            only_commits = set(only_commits)
        if reversed_order:
            logger.info("'reversed_order' is deprecated and will be removed in the next release. "
                        "Use 'order=reverse' instead. ")
            order = 'reverse'

        options = {
            "git_repo": None,
            "path_to_repo": path_to_repo,
            "from_commit": from_commit,
            "to_commit": to_commit,
            "from_tag": from_tag,
            "to_tag": to_tag,
            "since": since,
            "to": to,
            "single": single,
            "include_refs": include_refs,
            "include_remotes": include_remotes,
            "only_in_branch": only_in_branch,
            "only_modifications_with_file_types": only_modifications_with_file_types,
            "only_no_merge": only_no_merge,
            "only_authors": only_authors,
            "only_commits": only_commits,
            "only_releases": only_releases,
            "skip_whitespaces": skip_whitespaces,
            "filepath": filepath,
            "filepath_commits": None,
            "tagged_commits": None,
            "histogram": histogram_diff,
            "clone_repo_to": clone_repo_to,
            "order": order
        }
        self._conf = Conf(options)

    @staticmethod
    def _is_remote(repo: str) -> bool:
        return repo.startswith("git@") or repo.startswith("https://")

    def _clone_remote_repos(self, tmp_folder: str, repo: str) -> str:

        repo_folder = os.path.join(tmp_folder,
                                   self._get_repo_name_from_url(repo))
        logger.info("Cloning %s in temporary folder %s", repo, repo_folder)
        Repo.clone_from(url=repo, to_path=repo_folder)

        return repo_folder

    def _clone_folder(self) -> str:
        if self._conf.get('clone_repo_to'):
            clone_folder = str(Path(self._conf.get('clone_repo_to')))
            if not os.path.isdir(clone_folder):
                raise Exception("Not a directory: {0}".format(clone_folder))
        else:
            clone_folder = tempfile.TemporaryDirectory().name
        return clone_folder

    def traverse_commits(self) -> Generator[Commit, None, None]:
        """
        Analyze all the specified commits (all of them by default), returning
        a generator of commits.
        """
        for path_repo in self._conf.get('path_to_repos'):
            if self._is_remote(path_repo):
                path_repo = self._clone_remote_repos(self._clone_folder(), path_repo)

            git_repo = GitRepository(path_repo, self._conf)
            self._conf.set_value("git_repo", git_repo)
            self._conf.sanity_check_filters()

            logger.info('Analyzing git repository in %s', git_repo.path)

            # Get the commits that modified the filepath. In this case, we can not use
            # git rev-list since it doesn't have the option --follow, necessary to follow
            # the renames. Hence, we manually call git log instead
            if self._conf.get('filepath') is not None:
                self._conf.set_value('filepath_commits', git_repo.get_commits_modified_file(self._conf.get('filepath')))

            # Gets only the commits that are tagged
            if self._conf.get('only_releases'):
                self._conf.set_value('tagged_commits', git_repo.get_tagged_commits())

            # Build the arguments to pass to git rev-list.
            rev, kwargs = self._conf.build_args()

            # Iterate over all the commits returned by git rev-list
            for commit in git_repo.get_list_commits(rev, **kwargs):
                logger.info('Commit #%s in %s from %s', commit.hash, commit.committer_date, commit.author.name)

                if self._conf.is_commit_filtered(commit):
                    logger.info('Commit #%s filtered', commit.hash)
                    continue

                yield commit

            # cleaning, this is necessary since GitPython issues on memory leaks
            self._conf.set_value("git_repo", None)
            git_repo.clear()

    @staticmethod
    def _get_repo_name_from_url(url: str) -> str:
        last_slash_index = url.rfind("/")
        last_suffix_index = url.rfind(".git")
        if last_suffix_index < 0:
            last_suffix_index = len(url)

        if last_slash_index < 0 or last_suffix_index <= last_slash_index:
            raise Exception("Badly formatted url {}".format(url))

        return url[last_slash_index + 1:last_suffix_index]



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s', level=logging.INFO)

from pydriller.repository_mining import RepositoryMining


def test_between_revisions():
    from_tag = 'tag1'
    to_tag = 'tag3'

    lc = list(RepositoryMining('test-repos/tags',
                               from_tag=from_tag,
                               to_tag=to_tag).traverse_commits())

    assert len(lc) == 5
    assert '6bb9e2c6a8080e6b5b34e6e316c894b2ddbf7fcd' == lc[0].hash
    assert 'f1a90b8d7b151ceefd3e3dfc0dc1d0e12b5f48d0' == lc[1].hash
    assert '4638730126d40716e230c2040751a13153fb1556' == lc[2].hash
    assert 'a26f1438bd85d6b22497c0e5dae003812becd0bc' == lc[3].hash
    assert '627e1ad917a188a861c9fedf6e5858b79edbe439' == lc[4].hash



"""
Module that calculates the number of hunks made to a commit file.
"""
from statistics import median

from pydriller import ModificationType
from pydriller.metrics.process.process_metric import ProcessMetric


class HunksCount(ProcessMetric):
    """
    This class is responsible to implement the Number of Hunks metric for a
    file. As a hunk is a continuous block of changes in a diff, this number
    assesses how fragmented the commit file is (i.e. lots of changes all
    over the file versus one big change).

    If multiple commits are passed, it returns the median number of hunks in
    that range.
    """

    def count(self):
        """
        Return the number of hunks for each modified file.

        :return: int number of hunks
        """
        renamed_files = {}
        files = {}

        for commit in self.repo_miner.traverse_commits():

            for modified_file in commit.modifications:

                filepath = renamed_files.get(modified_file.new_path,
                                             modified_file.new_path)

                if modified_file.change_type == ModificationType.RENAME:
                    renamed_files[modified_file.old_path] = filepath

                diff = modified_file.diff
                is_hunk = False
                hunks = 0

                for line in diff.splitlines():
                    if line.startswith('+') or line.startswith('-'):
                        if not is_hunk:
                            is_hunk = True
                            hunks += 1
                    else:
                        is_hunk = False

                if filepath in files:
                    files[filepath].append(hunks)
                else:
                    files[filepath] = [hunks]

        for path, hunks in files.items():
            files[path] = median(hunks)

        return files



from pathlib import Path
from datetime import datetime

import pytest

from pydriller.metrics.process.commits_count import CommitsCount

TEST_DATA = [
    ('test-repos/pydriller', 'domain/developer.py', 'fdf671856b260aca058e6595a96a7a0fba05454b', 'ab36bf45859a210b0eae14e17683f31d19eea041', 2),
    ('test-repos/pydriller', 'domain/developer.py', 'ab36bf45859a210b0eae14e17683f31d19eea041', 'fdf671856b260aca058e6595a96a7a0fba05454b', 2)
]

@pytest.mark.parametrize('path_to_repo, filepath, from_commit, '
                         'to_commit, expected', TEST_DATA)
def test_with_commits(path_to_repo, filepath, from_commit, to_commit, expected):
    metric = CommitsCount(path_to_repo=path_to_repo,
                          from_commit=from_commit,
                          to_commit=to_commit)
    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected


TEST_DATA = [
    ('test-repos/pydriller', 'domain/developer.py', datetime(2018, 3, 21), datetime(2018, 3, 23), 2)
]
@pytest.mark.parametrize('path_to_repo, filepath, since, '
                         'to, expected', TEST_DATA)
def test_with_dates(path_to_repo, filepath, since, to, expected):
    metric = CommitsCount(path_to_repo=path_to_repo,
                          since=since,
                          to=to)
    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected



import pytest

from datetime import datetime
from pydriller.metrics.process.process_metric import ProcessMetric

dt1 = datetime(2016, 10, 8, 17, 0, 0)
dt2 = datetime(2016, 10, 8, 17, 59, 0)

TEST_DATA = [
    ('test-repos/pydriller', None, dt2, None, '81ddf7e78d92f3aaa212d5924d1ae0ed1fd980e6'), # Test (if not since and not from_commit) 
    ('test-repos/pydriller', dt1, None, 'ab36bf45859a210b0eae14e17683f31d19eea041', None)  # Test (if not to and not to_commit) 
]

@pytest.mark.parametrize('path_to_repo, since, to, from_commit, to_commit', TEST_DATA)
def test_type_error(path_to_repo, since, to, from_commit, to_commit):

    with pytest.raises(TypeError):
        ProcessMetric(path_to_repo=path_to_repo,
                      since=since,
                      to=to,
                      from_commit=from_commit,
                      to_commit=to_commit)



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s',
                    level=logging.INFO)

from pydriller.repository_mining import RepositoryMining

from datetime import datetime, timezone, timedelta


def test_one_timezone():
    lc = list(
        RepositoryMining('test-repos/branches_merged',
                         single='29e929fbc5dc6a2e9c620069b24e2a143af4285f')
            .traverse_commits())

    to_zone = timezone(timedelta(hours=2))
    dt = datetime(2016, 4, 4, 13, 21, 25, tzinfo=to_zone)

    assert lc[0].author_date == dt


def test_between_dates_reversed():
    lc = list(
        RepositoryMining('test-repos/different_files',
                         single='375de7a8275ecdc0b28dc8de2568f47241f443e9')
            .traverse_commits())

    to_zone = timezone(timedelta(hours=-4))
    dt = datetime(2016, 10, 8, 17, 57, 49, tzinfo=to_zone)

    assert lc[0].author_date == dt



"""
Module that calculates the number of hunks made to a commit file.
"""
import statistics

from pydriller import ModificationType
from pydriller.metrics.process.process_metric import ProcessMetric

class CodeChurn(ProcessMetric):
    """
    This class is responsible to implement the Code Churn metric for a
    file.
    A code churn is the sum of (added lines - removed lines) across the
    analyzed commits.
    It allows to count for the:
    * total number of code churns - count();
    * maximum code churn for all commits - max();
    * average code churn per commit.
    """

    def __init__(self, path_to_repo: str,
                 since=None,
                 to=None,
                 from_commit: str = None,
                 to_commit: str = None):

        super().__init__(path_to_repo, since=since, to=to, from_commit=from_commit, to_commit=to_commit)
        self._initialize()

    def _initialize(self):

        renamed_files = {}
        self.files = {}

        for commit in self.repo_miner.traverse_commits():

            for modified_file in commit.modifications:

                filepath = renamed_files.get(modified_file.new_path, modified_file.new_path)

                if modified_file.change_type == ModificationType.RENAME:
                    renamed_files[modified_file.old_path] = filepath

                churn = modified_file.added - modified_file.removed
                self.files.setdefault(filepath, []).append(churn)

    def count(self):
        """
        Return the total number of code churns for each modified file.

        :return: int number of churns
        """
        count = dict()
        for path, churns in self.files.items():
            count[path] = sum(churns)

        return count

    def max(self):
        """
        Return the maximum code churn for each modified file.

        :return: int max number of churns
        """
        max_count = dict()
        for path, churns in self.files.items():
            max_count[path] = max(churns)

        return max_count

    def avg(self):
        """
        Return the average number of code churns for each modified file.

        :return: int avg number of churns rounded off to the nearest integer
        """
        avg_count = dict()
        for path, churns in self.files.items():
            avg_count[path] = round(statistics.mean(churns))

        return avg_count



from pathlib import Path
from datetime import datetime

import pytest

from pydriller.metrics.process.contributors_experience import \
    ContributorsExperience

TEST_DATA = [
   ('test-repos/pydriller', 'domain/modification.py', 'fdf671856b260aca058e6595a96a7a0fba05454b', 'ab36bf45859a210b0eae14e17683f31d19eea041', 100.0),
   ('test-repos/pydriller', 'pydriller/git_repository.py', 'e9854bbea1cb7b7f06cbb559f7b06724d11ae1e5', 'e9854bbea1cb7b7f06cbb559f7b06724d11ae1e5', 100.0),
   ('test-repos/pydriller', 'pydriller/git_repository.py', 'e9854bbea1cb7b7f06cbb559f7b06724d11ae1e5', '9d0924301e4fae00eea6d00945bf834455e9a2a6', round(100*28/30, 2))
]

@pytest.mark.parametrize('path_to_repo, filepath, from_commit, to_commit, expected', TEST_DATA)
def test_with_commits(path_to_repo, filepath, from_commit, to_commit, expected):
    metric = ContributorsExperience(path_to_repo=path_to_repo,
                                    from_commit=from_commit,
                                    to_commit=to_commit)
    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected


TEST_DATA = [
   ('test-repos/pydriller', 'domain/modification.py', datetime(2018, 3, 21), datetime(2018, 3, 23), 100.0),
   ('test-repos/pydriller', 'pydriller/git_repository.py', datetime(2018, 8, 1), datetime(2018, 8, 2), 100.0),
   ('test-repos/pydriller', 'pydriller/git_repository.py',  datetime(2018, 7, 23), datetime(2018, 8, 2), round(100*28/30, 2))
]

@pytest.mark.parametrize('path_to_repo, filepath, since, to, expected', TEST_DATA)
def test_with_dates(path_to_repo, filepath, since, to, expected):
    metric = ContributorsExperience(path_to_repo=path_to_repo,
                                    since=since,
                                    to=to)
    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected


from pathlib import Path
from datetime import datetime

import pytest

from pydriller.metrics.process.code_churn import CodeChurn

TEST_DATA = [
    ('test-repos/pydriller', 'domain/commit.py', 'ab36bf45859a210b0eae14e17683f31d19eea041', '71e053f61fc5d31b3e31eccd9c79df27c31279bf', 47, 34, 16)
]

@pytest.mark.parametrize('path_to_repo, filepath, from_commit, to_commit, expected_count, expected_max, expected_avg', TEST_DATA)
def test_with_commits(path_to_repo, filepath, from_commit, to_commit, expected_count, expected_max, expected_avg):
    metric = CodeChurn(path_to_repo=path_to_repo,
                       from_commit=from_commit,
                       to_commit=to_commit)

    actual_count = metric.count()
    actual_max = metric.max()
    actual_avg = metric.avg()

    filepath = str(Path(filepath))

    assert actual_count[filepath] == expected_count
    assert actual_max[filepath] == expected_max
    assert actual_avg[filepath] == expected_avg


TEST_DATA = [
    ('test-repos/pydriller', 'domain/commit.py', datetime(2018, 3, 21), datetime(2018, 3, 27), 47, 34, 16)
]

@pytest.mark.parametrize('path_to_repo, filepath, since, to, expected_count, expected_max, expected_avg', TEST_DATA)
def test_with_dates(path_to_repo, filepath, since, to, expected_count, expected_max, expected_avg):
    metric = CodeChurn(path_to_repo=path_to_repo,
                       since=since,
                       to=to)

    actual_count = metric.count()
    actual_max = metric.max()
    actual_avg = metric.avg()

    filepath = str(Path(filepath))

    assert actual_count[filepath] == expected_count
    assert actual_max[filepath] == expected_max
    assert actual_avg[filepath] == expected_avg




# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import datetime, timezone, timedelta
from pydriller.repository_mining import RepositoryMining
from git import Repo

import pytest

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s',
                    level=logging.INFO)


def test_mod_with_file_types():
    lc = list(RepositoryMining('test-repos/different_files',
                               only_modifications_with_file_types=['.java']).traverse_commits())

    assert len(lc) == 2
    assert lc[0].hash == 'a1b6136f978644ff1d89816bc0f2bd86f6d9d7f5'
    assert lc[1].hash == 'b8c2be250786975f1c6f47e96922096f1bb25e39'

    lc = list(RepositoryMining('test-repos/different_files1',
                               only_modifications_with_file_types=['.java'])
              .traverse_commits())

    assert len(lc) == 2
    assert lc[0].hash == '5adbb71167e79ab6b974827e74c9da4d81977655'
    assert lc[1].hash == '0577bec2387ee131e1ccf336adcc172224d3f6f9'


def test_mod_with_file_types_no_extension():
    lc = list(RepositoryMining('test-repos/different_files',
                               only_modifications_with_file_types=['.py'])
              .traverse_commits())

    assert len(lc) == 0


def test_mod_with_file_types_and_date():
    to_zone = timezone(timedelta(hours=2))
    dt1 = datetime(2016, 10, 8, 23, 57, 49, tzinfo=to_zone)
    print(dt1)
    lc = list(RepositoryMining('test-repos/different_files',
                               only_modifications_with_file_types=['.java'],
                               since=dt1)
              .traverse_commits())

    print(lc)
    assert len(lc) == 1
    assert lc[0].hash == 'b8c2be250786975f1c6f47e96922096f1bb25e39'


def test_only_in_main_branch():
    lc = list(RepositoryMining('test-repos/branches_not_merged').traverse_commits())

    assert len(lc) == 3
    assert lc[0].hash == '04b0af7b53c2a0095e98951571aa41c2e0e0dbec'
    assert lc[1].hash == 'e51421e0beae6a3c20bdcdfc21066e05db675e03'
    assert lc[2].hash == 'b197ef4f0b4bc5b7d55c8949ecb1c861731f0b9d'


def test_only_no_merge():
    lc = list(RepositoryMining('test-repos/branches_merged',
                               only_no_merge=True).traverse_commits())

    assert len(lc) == 3
    assert lc[0].hash == '168b3aab057ed61a769acf336a4ef5e64f76c9fd'
    assert lc[1].hash == '8169f76a3d7add54b4fc7bca7160d1f1eede6eda'
    assert lc[2].hash == '8986af2a679759e5a15794f6d56e6d46c3f302f1'


def test_no_filters():
    lc = list(RepositoryMining('test-repos/different_files').traverse_commits())

    assert len(lc) == 3
    assert lc[0].hash == 'a1b6136f978644ff1d89816bc0f2bd86f6d9d7f5'
    assert lc[1].hash == '375de7a8275ecdc0b28dc8de2568f47241f443e9'
    assert lc[2].hash == 'b8c2be250786975f1c6f47e96922096f1bb25e39'


def test_only_in_branch():
    lc = list(RepositoryMining('test-repos/branches_not_merged',
                               only_in_branch='b1').traverse_commits())
    assert len(lc) == 5

    assert lc[0].hash == '04b0af7b53c2a0095e98951571aa41c2e0e0dbec'
    assert lc[1].hash == 'e51421e0beae6a3c20bdcdfc21066e05db675e03'
    assert lc[2].hash == 'b197ef4f0b4bc5b7d55c8949ecb1c861731f0b9d'
    assert lc[3].hash == '87a31153090808f1e6f679a14ea28729a0b74f4d'
    assert lc[4].hash == '702d469710d2087e662c210fd0e4f9418ec813fd'


def test_only_in_branches():
    # by default, only analyze master
    assert len(list(RepositoryMining('test-repos/branches_not_merged')
                    .traverse_commits())) == 3
    # only analyze b2
    assert len(list(RepositoryMining('test-repos/branches_not_merged',
                                     only_in_branch='b2')
                    .traverse_commits())) == 4
    # only analyze b1
    assert len(list(RepositoryMining('test-repos/branches_not_merged',
                                     only_in_branch='b1')
                    .traverse_commits())) == 5


def test_only_in_branch_not_exist():
    with pytest.raises(Exception):
        list(RepositoryMining('test-repos/branches_not_merged', only_in_branch='branch2')
             .traverse_commits())


def test_only_authors():
    lc = list(RepositoryMining('test-repos/multiple_authors',
                               only_authors=["Maurício Aniche"])
              .traverse_commits())
    assert len(lc) == 4

    lc = list(RepositoryMining('test-repos/multiple_authors',
                               only_authors=["ishepard"])
              .traverse_commits())
    assert len(lc) == 1


def test_only_authors_not_existing():
    lc = list(RepositoryMining('test-repos/multiple_authors',
                               only_authors=["Uncle Bob"])
              .traverse_commits())
    assert len(lc) == 0


def test_only_commits():
    lc = list(RepositoryMining('test-repos/complex_repo',
                               only_commits=["9e71dd5726d775fb4a5f08506a539216e878adbb"]).traverse_commits())
    assert len(lc) == 1
    assert lc[0].hash == "9e71dd5726d775fb4a5f08506a539216e878adbb"

    lc = list(RepositoryMining('test-repos/complex_repo',
                               only_commits=["953737b199de233896f00b4d87a0bc2794317253",
                                             "ffccf1e7497eb8136fd66ed5e42bef29677c4b71"]).traverse_commits())
    assert len(lc) == 2
    assert lc[0].hash == "ffccf1e7497eb8136fd66ed5e42bef29677c4b71"
    assert lc[1].hash == "953737b199de233896f00b4d87a0bc2794317253"

    lc = list(RepositoryMining('test-repos/complex_repo',
                               only_commits=["866e997a9e44cb4ddd9e00efe49361420aff2559",
                                             "57dbd017d1a744b949e7ca0b1c1a3b3dd4c1cbc1",
                                             "e7d13b0511f8a176284ce4f92ed8c6e8d09c77f2"]).traverse_commits())
    assert len(lc) == 3
    assert lc[0].hash == "866e997a9e44cb4ddd9e00efe49361420aff2559"
    assert lc[1].hash == "57dbd017d1a744b949e7ca0b1c1a3b3dd4c1cbc1"
    assert lc[2].hash == "e7d13b0511f8a176284ce4f92ed8c6e8d09c77f2"

    lc = list(RepositoryMining('test-repos/complex_repo',
                               only_commits=["fake hash"]).traverse_commits())
    assert len(lc) == 0

    total_commits = len(list(RepositoryMining('test-repos/complex_repo').traverse_commits()))

    assert total_commits == 13


def test_single_commit():
    lc = list(RepositoryMining('test-repos/complex_repo',
                               single="866e997a9e44cb4ddd9e00efe49361420aff2559").traverse_commits())
    assert len(lc) == 1
    assert lc[0].hash == "866e997a9e44cb4ddd9e00efe49361420aff2559"

    lc = list(RepositoryMining('test-repos/complex_repo',
                               single="ffccf1e7497eb8136fd66ed5e42bef29677c4b71").traverse_commits())
    assert len(lc) == 1
    assert lc[0].hash == "ffccf1e7497eb8136fd66ed5e42bef29677c4b71"


def test_single_commit_head():
    lc = list(RepositoryMining('test-repos/complex_repo',
                               single="e7d13b0511f8a176284ce4f92ed8c6e8d09c77f2").traverse_commits())
    assert len(lc) == 1

    lc_head = list(RepositoryMining('test-repos/complex_repo', single="HEAD").traverse_commits())
    assert len(lc_head) == 1
    assert lc[0].hash == lc_head[0].hash


def test_single_commit_not_existing_single_commit():
    with pytest.raises(Exception):
        list(RepositoryMining('test-repos/complex_repo',
                              single="ASD").traverse_commits())


def test_single_commit_not_existing_from_commit():
    with pytest.raises(Exception):
        list(RepositoryMining('test-repos/complex_repo',
                              from_commit="ASD").traverse_commits())


def test_single_commit_not_existing_to_commit():
    with pytest.raises(Exception):
        list(RepositoryMining('test-repos/complex_repo',
                              to_commit="ASD").traverse_commits())


def test_single_commit_not_existing_from_and_to_commit():
    with pytest.raises(Exception):
        list(RepositoryMining('test-repos/complex_repo',
                              from_commit="ASD",
                              to_commit="ASD").traverse_commits())


def test_filepath_with_to():
    dt = datetime(2018, 6, 6)
    assert len(list(RepositoryMining(
        path_to_repo='test-repos/szz',
        filepath='myfolder/A.java',
        to=dt).traverse_commits())) == 4


def test_filepath_with_since():
    since = datetime(2018, 6, 6)
    assert len(list(RepositoryMining(
        path_to_repo='test-repos/szz',
        filepath='myfolder/A.java',
        since=since).traverse_commits())) == 10


def test_filepath_with_rename():
    dt = datetime(2018, 6, 6)
    commits = list(RepositoryMining(
        path_to_repo='test-repos/small_repo',
        filepath='file4.java',
        to=dt).traverse_commits())
    assert len(commits) == 2

    commit_hashes = [commit.hash for commit in commits]

    assert 'da39b1326dbc2edfe518b90672734a08f3c13458' in commit_hashes
    assert 'a88c84ddf42066611e76e6cb690144e5357d132c' in commit_hashes


def test_filepath_with_rename_complex():
    commits = list(RepositoryMining(
        path_to_repo='test-repos/complex_repo',
        filepath='Matricula.javax').traverse_commits())
    assert len(commits) == 6

    commit_hashes = [commit.hash for commit in commits]

    assert 'f0dd1308bd904a9b108a6a40865166ee962af3d4' in commit_hashes
    assert '953737b199de233896f00b4d87a0bc2794317253' in commit_hashes
    assert 'a3290ac2f555eabca9e31180cf38e91f9e7e2761' in commit_hashes
    assert '71535a31f0b598a5d5fcebda7146ebc01def783a' in commit_hashes
    assert '57dbd017d1a744b949e7ca0b1c1a3b3dd4c1cbc1' in commit_hashes
    assert '866e997a9e44cb4ddd9e00efe49361420aff2559' in commit_hashes


def test_only_releases():
    lc = list(RepositoryMining('test-repos/tags',
                               only_releases=True).traverse_commits())

    assert len(lc) == 3
    assert '6bb9e2c6a8080e6b5b34e6e316c894b2ddbf7fcd' == lc[0].hash
    assert '4638730126d40716e230c2040751a13153fb1556' == lc[1].hash
    assert '627e1ad917a188a861c9fedf6e5858b79edbe439' == lc[2].hash


def test_only_releases_wo_releases():
    lc = list(RepositoryMining('test-repos/complex_repo', only_releases=True).traverse_commits())

    assert len(lc) == 0


def test_date_order():
    date_order = list(RepositoryMining('test-repos/order', order='date-order').traverse_commits())
    author_date_order = list(RepositoryMining('test-repos/order', order='author-date-order').traverse_commits())
    assert '5e3cfa27b3fe6dd4d12fd89664fea9397141b843' == date_order[0].hash == author_date_order[0].hash
    assert '19732de9e2b58ba7285f272810a9d8ddf18e7c89' == date_order[1].hash == author_date_order[1].hash
    assert '78a94953a3e140f2d0027fb57963345fbf6d59fe' == date_order[2].hash == author_date_order[2].hash
    assert '9cc3af5f242a1eba297f270acbdb8b6628556413' == date_order[3].hash == author_date_order[3].hash
    assert '6564f9e0bfb38725ebcfb4547e98e7f545c7de12' == date_order[4].hash == author_date_order[4].hash
    assert '5c95c1c6ba95a1bdb12772d1a63c7d331e664819' == date_order[5].hash == author_date_order[5].hash
    assert 'd23d7f6d37fd1163022a5dd46985acd34e6818d7' == date_order[6].hash == author_date_order[6].hash
    assert 'a45c8649b00d8b48cee04a822bd1d82acd667db2' == date_order[7].hash == author_date_order[7].hash


def test_topo_order():
    topo_order = list(RepositoryMining('test-repos/order', order='topo-order').traverse_commits())

    assert '5e3cfa27b3fe6dd4d12fd89664fea9397141b843' == topo_order[0].hash
    assert '19732de9e2b58ba7285f272810a9d8ddf18e7c89' == topo_order[1].hash
    assert '9cc3af5f242a1eba297f270acbdb8b6628556413' == topo_order[2].hash
    assert 'd23d7f6d37fd1163022a5dd46985acd34e6818d7' == topo_order[3].hash
    assert '78a94953a3e140f2d0027fb57963345fbf6d59fe' == topo_order[4].hash
    assert '6564f9e0bfb38725ebcfb4547e98e7f545c7de12' == topo_order[5].hash
    assert '5c95c1c6ba95a1bdb12772d1a63c7d331e664819' == topo_order[6].hash
    assert 'a45c8649b00d8b48cee04a822bd1d82acd667db2' == topo_order[7].hash


def test_should_visit_ascendent_order():
    lc = list(RepositoryMining('test-repos/small_repo').traverse_commits())
    assert len(lc) == 5
    assert lc[0].hash == 'a88c84ddf42066611e76e6cb690144e5357d132c'
    assert lc[1].hash == '6411e3096dd2070438a17b225f44475136e54e3a'
    assert lc[2].hash == '09f6182cef737db02a085e1d018963c7a29bde5a'
    assert lc[3].hash == '1f99848edadfffa903b8ba1286a935f1b92b2845'
    assert lc[4].hash == 'da39b1326dbc2edfe518b90672734a08f3c13458'


def test_should_visit_descendent_order():
    lc = list(RepositoryMining('test-repos/small_repo',
                               order='reverse').traverse_commits())
    assert len(lc) == 5
    assert lc[0].hash == 'da39b1326dbc2edfe518b90672734a08f3c13458'
    assert lc[1].hash == '1f99848edadfffa903b8ba1286a935f1b92b2845'
    assert lc[2].hash == '09f6182cef737db02a085e1d018963c7a29bde5a'
    assert lc[3].hash == '6411e3096dd2070438a17b225f44475136e54e3a'
    assert lc[4].hash == 'a88c84ddf42066611e76e6cb690144e5357d132c'


def test_should_visit_descendent_order_with_filters():
    lc = list(RepositoryMining('test-repos/small_repo',
                               from_commit='1f99848edadfffa903b8ba1286a935f1b92b2845',
                               to_commit='6411e3096dd2070438a17b225f44475136e54e3a',
                               order='reverse').traverse_commits())
    assert len(lc) == 3
    assert lc[0].hash == '1f99848edadfffa903b8ba1286a935f1b92b2845'
    assert lc[1].hash == '09f6182cef737db02a085e1d018963c7a29bde5a'
    assert lc[2].hash == '6411e3096dd2070438a17b225f44475136e54e3a'


def test_should_visit_descendent_order_with_filters_reversed():
    lc = list(RepositoryMining('test-repos/small_repo',
                               from_commit='6411e3096dd2070438a17b225f44475136e54e3a',
                               to_commit='1f99848edadfffa903b8ba1286a935f1b92b2845',
                               order='reverse').traverse_commits())
    assert len(lc) == 3
    assert lc[0].hash == '1f99848edadfffa903b8ba1286a935f1b92b2845'
    assert lc[1].hash == '09f6182cef737db02a085e1d018963c7a29bde5a'
    assert lc[2].hash == '6411e3096dd2070438a17b225f44475136e54e3a'


def test_include_refs():
    commits_no_refs = list(RepositoryMining('test-repos/branches_not_merged/',
                                            include_refs=False).traverse_commits())
    assert len(commits_no_refs) == 3
    commit_no_refs_hashes = [commit.hash for commit in commits_no_refs]

    commits_with_refs = list(RepositoryMining('test-repos/branches_not_merged/',
                                              include_refs=True).traverse_commits())
    assert len(commits_with_refs) == 6
    commit_with_refs_hashes = [commit.hash for commit in commits_with_refs]

    commits_not_in_commits_no_refs = list(set(commit_with_refs_hashes) - set(commit_no_refs_hashes))
    assert len(commits_not_in_commits_no_refs) == 3

    # First commit on branch b1
    assert '87a31153090808f1e6f679a14ea28729a0b74f4d' in commits_not_in_commits_no_refs
    # Commit that branch b1 points to
    assert '702d469710d2087e662c210fd0e4f9418ec813fd' in commits_not_in_commits_no_refs
    # Commit that branch b2 points to
    assert '7203c0b8220dcc7a59614bc7549799cd203ac072' in commits_not_in_commits_no_refs


def test_include_remotes():
    repo = Repo('test-repos/pydriller/')

    # Prove that a recent commit is not in 'test-repos/pydriller'
    commits_before_fetch = [commit.hash for commit in RepositoryMining('test-repos/pydriller/').traverse_commits()]
    assert '2fa0d8c57829b086c9372722115a89d11c9bdd35' not in commits_before_fetch

    # Fetch and re-assert that the commit is still not in the repo
    repo.git.fetch('--all')
    commits_after_fetch = [commit.hash for commit in RepositoryMining('test-repos/pydriller/').traverse_commits()]
    assert '2fa0d8c57829b086c9372722115a89d11c9bdd35' not in commits_after_fetch

    # Set include_remotes=True and assert that the commit is now found
    commits_with_remotes = [commit.hash for commit in RepositoryMining('test-repos/pydriller/', include_remotes=True).traverse_commits()]
    assert '2fa0d8c57829b086c9372722115a89d11c9bdd35' in commits_with_remotes



"""
Module that calculates the number of normalized added and deleted lines of a
file.
"""
import statistics
from pydriller import ModificationType
from pydriller.metrics.process.process_metric import ProcessMetric


class LinesCount(ProcessMetric):
    """
    This class is responsible to implement the following metrics:

    * Changed Lines: the number of added and deleted lines in the evolution
        period [from_commit, to_commit]

    * Added Lines: the sum over all commits of the lines of code added to a
        file in the evolution period [from_commit, to_commit]

    * Max Added Lines: the maximum number of lines of code added to a file
        per commit in the evolution period [from_commit, to_commit]

    * Average Added Lines: the average lines of code added to a file per commit
        in the evolution period [from_commit, to_commit]

    * Removed Lines: the sum over all commits of the lines of code removed to a
        file in the evolution period [from_commit, to_commit]

    * Max Removed Lines: the maximum number of lines of code removed to a file
        per commit in the evolution period [from_commit, to_commit]

    * Average Removed Lines: the average lines of code removed to a file per
        commit in the evolution period [from_commit, to_commit]

    """

    def __init__(self, path_to_repo: str,
                 since=None,
                 to=None,
                 from_commit: str = None,
                 to_commit: str = None):

        super().__init__(path_to_repo, since=since, to=to, from_commit=from_commit, to_commit=to_commit)
        self._initialize()

    def _initialize(self):

        self.lines_added = dict()
        self.lines_removed = dict()

        renamed_files = {}
        for commit in self.repo_miner.traverse_commits():

            for modified_file in commit.modifications:

                filepath = renamed_files.get(modified_file.new_path,
                                             modified_file.new_path)

                if modified_file.change_type == ModificationType.RENAME:
                    renamed_files[modified_file.old_path] = filepath

                self.lines_added.setdefault(filepath, []).append(modified_file.added)
                self.lines_removed.setdefault(filepath, []).append(modified_file.removed)

    def count(self):
        """
        Sum over all commits of the lines of code added and removed to a file .

        :return: int lines added + lines removed
        """
        count = dict()

        for path, lines in self.lines_added.items():
            count[path] = count.get(path, 0) + sum(lines)

        for path, lines in self.lines_removed.items():
            count[path] = count.get(path, 0) + sum(lines)

        return count

    def count_added(self):
        """
        Sum over all commits of the lines of code added to a file .

        :return: int lines added
        """
        count = dict()
        for path, lines in self.lines_added.items():
            count[path] = sum(lines)

        return count

    def max_added(self):
        """
        Maximum number of lines of code added to a file for all commits

        :return: int max number of lines added
        """
        maximum = dict()
        for path, lines in self.lines_added.items():
            maximum[path] = max(lines)

        return maximum

    def avg_added(self):
        """
        Average lines of code added to a file per commit

        :return: int avg number of lines rounded off to the nearest integer
        """
        avg = dict()
        for path, lines in self.lines_added.items():
            avg[path] = round(statistics.mean(lines))

        return avg

    def count_removed(self):
        """
        Sum over all commits of the lines of code removed to a file .

        :return: int lines removed
        """
        count = dict()
        for path, lines in self.lines_removed.items():
            count[path] = sum(lines)

        return count

    def max_removed(self):
        """
        Maximum number of lines of code removed in a file for all commits

        :return: int max number of lines removed
        """
        maximum = dict()
        for path, lines in self.lines_removed.items():
            maximum[path] = max(lines)

        return maximum

    def avg_removed(self):
        """
        Average lines of code removed in a file per commit

        :return: int rounded off to the nearest integer
        """
        avg = dict()
        for path, lines in self.lines_removed.items():
            avg[path] = round(statistics.mean(lines))

        return avg



"""
Configuration module.
"""

import logging
from datetime import datetime

import pytz
from gitdb.exc import BadName

from pydriller.domain.commit import Commit

logger = logging.getLogger(__name__)


class Conf:
    """
    Configuration class. This class holds all the possible configurations of
    the mining process (i.e., starting and ending dates, branches, etc.)
    It's also responsible for checking whether the filters are correct (i.e.,
    the user did not specify 2 starting commits).
    """

    def __init__(self, options):
        # insert all the configurations in a local dictionary
        self._options = {}
        for key, val in options.items():
            self._options[key] = val

        self._sanity_check_repos(self.get('path_to_repo'))
        if isinstance(self.get('path_to_repo'), str):
            self.set_value('path_to_repos', [self.get('path_to_repo')])
        else:
            self.set_value('path_to_repos', self.get('path_to_repo'))

    def set_value(self, key, value):
        """
        Save the value of a configuration.

        :param key: configuration (i.e., start date)
        :param value: value
        """
        self._options[key] = value

    def get(self, key):
        """
        Return the value of the configuration.

        :param key: configuration name
        :return: value of the configuration, None if not present
        """
        return self._options.get(key, None)

    @staticmethod
    def _sanity_check_repos(path_to_repo):
        """
        Checks if repo is of type str or list.

        @param path_to_repo: path to the repo as provided by the user.
        @return:
        """
        if not isinstance(path_to_repo, str) and \
                not isinstance(path_to_repo, list):
            raise Exception("The path to the repo has to be of type 'string' or 'list of strings'!")

    def sanity_check_filters(self):
        """
        Check if the values passed by the user are correct.

        """
        self._check_correct_filters_order()
        self.check_starting_commit()
        self.check_ending_commit()
        self._check_timezones()

        # Check if from_commit and to_commit point to the same commit, in which case
        # we remove both filters and use the "single" filter instead. This prevents
        # errors with dates.
        if self.get("from_commit") and self.get("to_commit") and self.get("from_commit") == self.get("to_commit"):
            logger.warning("You should not point from_commit and to_commit to the same "
                           "commit, but use the 'single' filter instead.")
            single = self.get("to_commit")
            self.set_value("from_commit", None)
            self.set_value("to_commit", None)
            self.set_value("single", single)

        if self.get('single') is not None:
            if any([self.get('since'),
                    self.get('to'),
                    self.get('from_commit'),
                    self.get('to_commit'),
                    self.get('from_tag'),
                    self.get('to_tag')]):
                raise Exception('You can not specify a single commit with '
                                'other filters')
            try:
                self.set_value('single', self.get("git_repo").get_commit(
                    self.get('single')).hash)
            except BadName:
                raise Exception("The commit {} defined in "
                                "the 'single' filtered does "
                                "not exist".format(self.get('single')))

    def _check_correct_filters_order(self):
        """
        Check that from_commit comes before to_commit
        """
        if self.get('from_commit') and self.get('to_commit'):
            chronological_order = self._is_commit_before(
                self.get('git_repo').get_commit(self.get('from_commit')),
                self.get('git_repo').get_commit(self.get('to_commit')))

            if not chronological_order:
                self._swap_commit_fiters()

    def _swap_commit_fiters(self):
        # reverse from and to commit
        from_commit = self.get('from_commit')
        to_commit = self.get('to_commit')
        self.set_value('from_commit', to_commit)
        self.set_value('to_commit', from_commit)

    @staticmethod
    def _is_commit_before(commit_before: Commit, commit_after: Commit):
        if commit_before.committer_date < commit_after.committer_date:
            return True
        if commit_before.committer_date == commit_after.committer_date and \
                commit_before.author_date < commit_after.author_date:
            return True
        return False

    def check_starting_commit(self):
        """
        Get the starting commit from the 'since', 'from_commit' or 'from_tag'
        filter.
        """
        if not self.only_one_filter([self.get('since'),
                                     self.get('from_commit'),
                                     self.get('from_tag')]):
            raise Exception('You can only specify one between since, from_tag and from_commit')
        if self.get('from_tag') is not None:
            tagged_commit = self.get("git_repo").get_commit_from_tag(self.get('from_tag'))
            self.set_value('from_commit', tagged_commit.hash)
        if self.get('from_commit'):
            try:
                commit = self.get("git_repo").get_commit(self.get('from_commit'))
                if len(commit.parents) == 0:
                    self.set_value('from_commit', [commit.hash])
                elif len(commit.parents) == 1:
                    self.set_value('from_commit', ['^' + commit.hash + '^'])
                else:
                    commits = ['^' + x for x in commit.parents]
                    self.set_value('from_commit', commits)
            except Exception:
                raise Exception("The commit {} defined in the 'from_tag' or 'from_commit' filter does "
                                "not exist".format(self.get('from_commit')))

    def check_ending_commit(self):
        """
        Get the ending commit from the 'to', 'to_commit' or 'to_tag' filter.
        """
        if not self.only_one_filter([self.get('to'),
                                     self.get('to_commit'),
                                     self.get('to_tag')]):
            raise Exception('You can only specify one between since, from_tag and from_commit')
        if self.get('to_tag') is not None:
            tagged_commit = self.get("git_repo").get_commit_from_tag(self.get('to_tag'))
            self.set_value('to_commit', tagged_commit.hash)
        if self.get('to_commit'):
            try:
                commit = self.get("git_repo").get_commit(self.get('to_commit'))
                self.set_value('to_commit', commit.hash)
            except Exception:
                raise Exception("The commit {} defined in the 'to_tag' or 'to_commit' filter does "
                                "not exist".format(self.get('to_commit')))

    @staticmethod
    def only_one_filter(arr):
        """
        Return true if in 'arr' there is at most 1 filter to True.

        :param arr: iterable object
        :return:
        """
        return len([x for x in arr if x is not None]) <= 1

    def build_args(self):
        """
        This function builds the argument for git rev-list.

        :return:
        """
        single = self.get('single')
        since = self.get('since')
        until = self.get('to')
        from_commit = self.get('from_commit')
        to_commit = self.get('to_commit')
        include_refs = self.get('include_refs')
        remotes = self.get('include_remotes')
        branch = self.get('only_in_branch')
        authors = self.get('only_authors')
        order = self.get('order')
        rev = []
        kwargs = {}

        if single is not None:
            rev = [single, '-n', 1]
        elif from_commit is not None or to_commit is not None:
            if from_commit is not None and to_commit is not None:
                rev.extend(from_commit)
                rev.append(to_commit)
            elif from_commit is not None:
                rev.extend(from_commit)
                rev.append('HEAD')
            else:
                rev = to_commit
        elif branch is not None:
            rev = branch
        else:
            rev = 'HEAD'

        if self.get('only_no_merge'):
            kwargs['no-merges'] = True

        if not order:
            kwargs['reverse'] = True
        elif order == 'reverse':
            kwargs['reverse'] = False
        elif order == 'date-order':
            kwargs['date-order'] = True
        elif order == 'author-date-order':
            kwargs['author-date-order'] = True
        elif order == 'topo-order':
            kwargs['topo-order'] = True

        if include_refs is not None:
            kwargs['all'] = include_refs

        if remotes is not None:
            kwargs['remotes'] = remotes

        if authors is not None:
            kwargs['author'] = authors

        if since is not None:
            kwargs['since'] = since

        if until is not None:
            kwargs['until'] = until

        return rev, kwargs

    def is_commit_filtered(self, commit: Commit):
        # pylint: disable=too-many-branches,too-many-return-statements
        """
        Check if commit has to be filtered according to the filters provided
        by the user.

        :param Commit commit: Commit to check
        :return:
        """
        if self.get('only_modifications_with_file_types') is not None:
            if not self._has_modification_with_file_type(commit):
                logger.debug('Commit filtered for modification types')
                return True
        if self.get('only_commits') is not None and commit.hash not in self.get('only_commits'):
            logger.debug("Commit filtered because it is not one of the specified commits")
            return True
        if self.get('filepath_commits') is not None and commit.hash not in self.get('filepath_commits'):
            logger.debug("Commit filtered because it did not modify the specified file")
            return True
        if self.get('tagged_commits') is not None and commit.hash not in self.get('tagged_commits'):
            logger.debug("Commit filtered because it is not tagged")
            return True
        return False

    def _has_modification_with_file_type(self, commit):
        for mod in commit.modifications:
            if mod.filename.endswith(tuple(self.get('only_modifications_with_file_types'))):
                return True
        return False

    def _check_timezones(self):
        if self.get('since') is not None:
            self.set_value('since', self._replace_timezone(self.get('since')))
        if self.get('to') is not None:
            self.set_value('to', self._replace_timezone(self.get('to')))

    @staticmethod
    def _replace_timezone(dt: datetime):
        if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
            dt = dt.replace(tzinfo=pytz.utc)
        return dt



"""
Module that calculates the number of files committed together.
"""
import statistics

from pydriller.metrics.process.process_metric import ProcessMetric

class ChangeSet(ProcessMetric):
    """
    This class is responsible to implement the Change Set metric that
    measures the

    * maximum number of files committed together - max();
    * average number of files committed together - avg().
    """

    def __init__(self, path_to_repo: str,
                 since=None,
                 to=None,
                 from_commit: str = None,
                 to_commit: str = None):

        super().__init__(path_to_repo, since=since, to=to, from_commit=from_commit, to_commit=to_commit)
        self._initialize()

    def _initialize(self):

        self.committed_together = []

        for commit in self.repo_miner.traverse_commits():
            self.committed_together.append(len(commit.modifications))

    def max(self):
        """
        Return the maximum number of files committed together.

        :return: int max number of files committed together
        """
        return max(self.committed_together, default=0)

    def avg(self):
        """
        Return the average number of files committed together.

        :return: int avg number of files rounded off to the nearest integer
        """
        if not self.committed_together:
            return 0

        return round(statistics.mean(self.committed_together))



"""
Module that calculates the History Complexity Period Factor (HCPF) \
for the History Complexity Metric (HCM).

The History Complexity Metric can be calculated by calling the \
method count as many times as the evolution period to analyze and \
summing up the results.

E.g.

hcpf_1 = HistoryComplexity(..., from_commit=c1, to_commit=c2).count()
hcpf_2 = HistoryComplexity(..., from_commit=c3, to_commit=c4).count()

hcm = hcpf_1 + hcpf_2

See https://ieeexplore.ieee.org/document/5070510
"""

from math import log

from pydriller import ModificationType
from pydriller.metrics.process.process_metric import ProcessMetric


class HistoryComplexity(ProcessMetric):
    """
    This class is responsible to implement the History Complexity Metric. \
    The metric assigns to each modified file the effect of the change \
    complexity of an evolution period.
    """

    def count(self):
        """
        Calculate the History Complexity Period Factor for each modified file \
        in the provided period [from_commit, to_commit], returning the \
        probability of each file being modified during that period

        :return: dict
        {
            filepath: float
        }
        """

        renamed_files = {}
        files = {}

        for commit in self.repo_miner.traverse_commits():

            for modified_file in commit.modifications:
                filepath = renamed_files.get(modified_file.new_path,
                                             modified_file.new_path)

                if modified_file.change_type == ModificationType.RENAME:
                    renamed_files[modified_file.old_path] = filepath

                modifications = modified_file.added + modified_file.removed
                if modifications:
                    files[filepath] = files.get(filepath, 0) + modifications

        # Total lines modified in the period
        total_modifications = sum(files.values())

        # Number of modified files in the period
        n_files = len(files)

        # From absolute number of changes to relative number of changes
        for filepath in files:
            files[filepath] /= total_modifications

        # Normalized entropy
        entropy = 0
        if len(files.values()) > 1:
            entropy = -sum(p*log(p+1/1e10, n_files) for p in files.values())

        for filepath in files:
            files[filepath] *= entropy
            files[filepath] = round(files[filepath] * 100, 2)

        return files



from pathlib import Path
from datetime import datetime

import pytest

from pydriller.metrics.process.hunks_count import HunksCount

TEST_DATA = [
    ('test-repos/pydriller', 'scm/git_repository.py', '71e053f61fc5d31b3e31eccd9c79df27c31279bf', '71e053f61fc5d31b3e31eccd9c79df27c31279bf', 8),
    ('test-repos/pydriller', 'scm/git_repository.py', 'ab36bf45859a210b0eae14e17683f31d19eea041', '71e053f61fc5d31b3e31eccd9c79df27c31279bf', 3),
    ('test-repos/pydriller', 'domain/modification.py', 'ab36bf45859a210b0eae14e17683f31d19eea041', 'fdf671856b260aca058e6595a96a7a0fba05454b', 1)
]

@pytest.mark.parametrize('path_to_repo, filepath, from_commit, to_commit, expected', TEST_DATA)
def test_with_commits(path_to_repo, filepath, from_commit, to_commit, expected):
    metric = HunksCount(path_to_repo=path_to_repo,
                        from_commit=from_commit,
                        to_commit=to_commit)

    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected

TEST_DATA = [
    ('test-repos/pydriller', 'scm/git_repository.py', datetime(2018, 3, 26), datetime(2018, 3, 27), 8),
    ('test-repos/pydriller', 'scm/git_repository.py', datetime(2018, 3, 21), datetime(2018, 3, 27), 3),
    ('test-repos/pydriller', 'domain/modification.py', datetime(2018, 3, 21), datetime(2018, 3, 22, 23), 1)
]

@pytest.mark.parametrize('path_to_repo, filepath, since, to, expected', TEST_DATA)
def test_with_dates(path_to_repo, filepath, since, to, expected):
    metric = HunksCount(path_to_repo=path_to_repo,
                        since=since,
                        to=to)

    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected


"""
Module that calculates the number of developers that contributed to each
modified file in the repo in a given time range.

See https://dl.acm.org/doi/10.1145/2025113.2025119
"""
from pydriller import ModificationType
from pydriller.metrics.process.process_metric import ProcessMetric


class ContributorsCount(ProcessMetric):
    """
    This class is responsible to implement the following metrics:

    * Contributors Count: measures the number of contributors who modified a
      file.

    * Minor Contributors Count: measures the number of contributors who
      authored less than 5% of code of a file.
    """

    def __init__(self, path_to_repo: str,
                 since=None,
                 to=None,
                 from_commit: str = None,
                 to_commit: str = None):

        super().__init__(path_to_repo, since=since, to=to, from_commit=from_commit, to_commit=to_commit)
        self._initialize()

    def _initialize(self):

        self.contributors = dict()
        self.minor_contributors = dict()

        renamed_files = {}

        for commit in self.repo_miner.traverse_commits():

            for modified_file in commit.modifications:

                filepath = renamed_files.get(modified_file.new_path,
                                             modified_file.new_path)

                if modified_file.change_type == ModificationType.RENAME:
                    renamed_files[modified_file.old_path] = filepath

                author = commit.author.email.strip()
                lines_authored = modified_file.added + modified_file.removed

                self.contributors[filepath] = self.contributors.get(filepath, {})
                self.contributors[filepath][author] = self.contributors[filepath].get(author, 0) + lines_authored

        for path, contributions in list(self.contributors.items()):
            total = sum(contributions.values())
            if total == 0:
                del self.contributors[path]
            else:
                contributors_count = len(contributions.values())
                minor_contributors_count = sum(1
                                               for v in contributions.values()
                                               if v/total < .05)

                self.contributors[path] = contributors_count
                self.minor_contributors[path] = minor_contributors_count

    def count(self):
        """
        Return the number of contributors who modified a file.
        """
        return self.contributors

    def count_minor(self):
        """
        Return the number of contributors that authored less than
        5% of code of a file.
        """
        return self.minor_contributors



"""
Module that calculates the experience of contributors of a file.
"""
from pydriller import ModificationType
from pydriller.metrics.process.process_metric import ProcessMetric


class ContributorsExperience(ProcessMetric):
    """
    This class is responsible to implement the metric to measure the
    percentage of the lines authored by the highest contributor of a
    file in the provided evolution period [from_commit, to_commit].
    """

    def count(self):
        """
        Return the percentage of the lines authored by the highest contributor
        of a file for each modified file in the repository in the provided
        time range [from_commit, to_commit]

        :return: dict { filepath: float }
        of number of contributors for each modified file
        """
        renamed_files = {}
        files = {}

        for commit in self.repo_miner.traverse_commits():

            for modified_file in commit.modifications:

                filepath = renamed_files.get(modified_file.new_path,
                                             modified_file.new_path)

                if modified_file.change_type == ModificationType.RENAME:
                    renamed_files[modified_file.old_path] = filepath

                author = commit.author.email.strip()
                lines_authored = modified_file.added + modified_file.removed

                files[filepath] = files.get(filepath, {})
                files[filepath][author] = files[filepath].get(author, 0) + \
                    lines_authored

        for path, contributions in list(files.items()):
            total = sum(contributions.values())
            if total == 0:
                del files[path]
            else:
                files[path] = round(100*max(contributions.values()) / total, 2)

        return files



from pathlib import Path
from datetime import datetime

import pytest

from pydriller.metrics.process.lines_count import LinesCount

TEST_DATA = [
   ('test-repos/pydriller', '.gitignore', 'ab36bf45859a210b0eae14e17683f31d19eea041', 'ab36bf45859a210b0eae14e17683f31d19eea041', 197),
   ('test-repos/pydriller', 'domain/modification.py', 'ab36bf45859a210b0eae14e17683f31d19eea041', '71e053f61fc5d31b3e31eccd9c79df27c31279bf', 65)
]

@pytest.mark.parametrize('path_to_repo, filepath, from_commit, to_commit, expected_count', TEST_DATA)
def test_with_commits(path_to_repo, filepath, from_commit, to_commit, expected_count):
    metric = LinesCount(path_to_repo=path_to_repo,
                        from_commit=from_commit,
                        to_commit=to_commit)

    actual_count = metric.count()
    filepath = str(Path(filepath))

    assert actual_count[filepath] == expected_count

TEST_DATA = [
   ('test-repos/pydriller', '.gitignore', datetime(2018, 3, 21), datetime(2018, 3, 22), 197),
   ('test-repos/pydriller', 'domain/modification.py', datetime(2018, 3, 21), datetime(2018, 3, 27), 65)
]

@pytest.mark.parametrize('path_to_repo, filepath, since, to, expected_count', TEST_DATA)
def test_with_dates(path_to_repo, filepath, since, to, expected_count):
    metric = LinesCount(path_to_repo=path_to_repo,
                        since=since,
                        to=to)

    actual_count = metric.count()
    filepath = str(Path(filepath))

    assert actual_count[filepath] == expected_count


from setuptools import setup, find_packages

with open('requirements.txt') as reqs_file:
    requirements = reqs_file.read().splitlines()

with open('test-requirements.txt') as reqs_file:
    test_requirements = reqs_file.read().splitlines()

# Get the long description from the relevant file
long_description = 'PyDriller is a Python framework that helps developers on ' \
                   'mining software repositories. With PyDriller' \
                   ' you can easily extract information from any Git ' \
                   'repository, such as commits, developers, ' \
                   'modifications, diffs, and source codes, and ' \
                   'quickly export CSV files.'

setup(
    name='PyDriller',
    description='Framework for MSR',
    long_description=long_description,
    author='Davide Spadini',
    author_email='spadini.davide@gmail.com',
    version='1.14',
    packages=find_packages('.'),
    url='https://github.com/ishepard/pydriller',
    license='Apache License',
    package_dir={'pydriller': 'pydriller'},
    python_requires='>=3.5',
    install_requires=requirements,
    test_requirements=requirements + test_requirements,
    classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Topic :: Software Development :: Libraries :: Python Modules',
            "Operating System :: OS Independent",
            "Operating System :: POSIX",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: MacOS :: MacOS X",
            ]
)



import logging
from datetime import datetime

import pytest

from pydriller import RepositoryMining, GitRepository

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s',
                    level=logging.INFO)


@pytest.fixture
def path():
    return None


@pytest.fixture
def to():
    return None


@pytest.fixture
def repo(path):
    return list(RepositoryMining(path_to_repo=path).traverse_commits())


@pytest.fixture
def repo_to(path, to):
    return list(RepositoryMining(path_to_repo=path, to=to).traverse_commits())


@pytest.fixture()
def git_repo(path):
    gr = GitRepository(path)
    yield gr
    gr.clear()


# It should fail when no URLs are specified
def test_no_url():
    with pytest.raises(Exception):
        list(RepositoryMining().traverse_commits())


# It should fail when URL is not a string or a List
def test_badly_formatted_repo_url():
    with pytest.raises(Exception):
        list(RepositoryMining(path_to_repo=set('repo')).traverse_commits())


@pytest.mark.parametrize('path,expected', [
    ("test-repos/small_repo", 5)
])
def test_simple_url(repo, expected):
    assert len(repo) == expected


@pytest.mark.parametrize('path,expected', [
    (["test-repos/small_repo", "test-repos/branches_merged"], 9)
])
def test_two_local_urls(repo, expected):
    assert len(repo) == expected


@pytest.mark.parametrize('path,to,expected', [
    ("https://github.com/ishepard/pydriller.git",
     datetime(2018, 10, 20),
     159)
])
def test_simple_remote_url(repo_to, expected):
    assert len(repo_to) == expected


@pytest.mark.parametrize('path,to,expected', [
    (["https://github.com/mauricioaniche/repodriller.git",
      "https://github.com/ishepard/pydriller"],
     datetime(2018, 10, 20),
     518)
])
def test_two_remote_urls(repo_to, expected):
    assert len(repo_to) == expected


@pytest.mark.parametrize('path,expected', [
    (["test-repos/small_repo", "test-repos/small_repo"], 10)
])
def test_2_identical_local_urls(repo, expected):
    assert len(repo) == expected


@pytest.mark.parametrize('path,to,expected', [
    (["test-repos/small_repo", "https://github.com/ishepard/pydriller.git"],
     datetime(2018, 10, 20),
     164)
])
def test_both_local_and_remote_urls(repo_to, expected):
    assert len(repo_to) == expected


@pytest.mark.parametrize('path,to,expected', [
    (["test-repos/small_repo", "https://github.com/mauricioaniche/repodriller.git",
      "test-repos/branches_merged", "https://github.com/ishepard/pydriller.git"],
     datetime(2018, 10, 20),
     527)
])
def test_both_local_and_remote_urls_list(repo_to, expected):
    assert len(repo_to) == expected


def test_badly_formatted_url():
    with pytest.raises(Exception):
        list(RepositoryMining(
            path_to_repo='https://github.com/ishepard.git/test')
             .traverse_commits())

    with pytest.raises(Exception):
        list(RepositoryMining(path_to_repo='test').traverse_commits())


@pytest.mark.parametrize('path', ["test-repos/histogram"])
def test_diff_without_histogram(git_repo):
    # without histogram
    commit = list(RepositoryMining('test-repos/histogram',
                                   single="93df8676e6fab70d9677e94fd0f6b17db095e890").traverse_commits())[0]

    diff = commit.modifications[0].diff_parsed
    assert len(diff['added']) == 11
    assert (3, '    if (path == null)') in diff['added']
    assert (5, '        log.error("Icon path is null");') in diff['added']
    assert (6, '        return null;') in diff['added']
    assert (8, '') in diff['added']
    assert (9, '    java.net.URL imgURL = GuiImporter.class.getResource(path);') in diff['added']
    assert (10, '') in diff['added']
    assert (11, '    if (imgURL == null)') in diff['added']
    assert (12, '    {') in diff['added']
    assert (14, '        return null;') in diff['added']
    assert (16, '    else') in diff['added']
    assert (17, '        return new ImageIcon(imgURL);') in diff['added']

    assert len(diff['deleted']) == 7
    assert (3, '    java.net.URL imgURL = GuiImporter.class.getResource(path);') in diff['deleted']
    assert (4, '') in diff['deleted']
    assert (5, '    if (imgURL != null)') in diff['deleted']
    assert (7, '        return new ImageIcon(imgURL);') in diff['deleted']
    assert (9, '    else') in diff['deleted']
    assert (10, '    {') in diff['deleted']
    assert (13, '    return null;') in diff['deleted']


@pytest.mark.parametrize('path', ["test-repos/histogram"])
def test_diff_with_histogram(git_repo):
    # with histogram
    commit = list(RepositoryMining('test-repos/histogram',
                                   single="93df8676e6fab70d9677e94fd0f6b17db095e890",
                                   histogram_diff=True).traverse_commits())[0]
    diff = commit.modifications[0].diff_parsed
    assert (4, '    {') in diff["added"]
    assert (5, '        log.error("Icon path is null");') in diff["added"]
    assert (6, '        return null;') in diff["added"]
    assert (7, '    }') in diff["added"]
    assert (8, '') in diff["added"]
    assert (11, '    if (imgURL == null)') in diff["added"]
    assert (12, '    {') in diff["added"]
    assert (13, '        log.error("Couldn\'t find icon: " + imgURL);') in diff["added"]
    assert (14, '        return null;') in diff["added"]
    assert (17, '        return new ImageIcon(imgURL);') in diff["added"]

    assert (6, '    {') in diff["deleted"]
    assert (7, '        return new ImageIcon(imgURL);') in diff["deleted"]
    assert (10, '    {') in diff["deleted"]
    assert (11, '        log.error("Couldn\'t find icon: " + imgURL);') in diff["deleted"]
    assert (12, '    }') in diff["deleted"]
    assert (13, '    return null;') in diff["deleted"]


def test_ignore_add_whitespaces():
    commit = list(RepositoryMining('test-repos/whitespace',
                                   single="338a74ceae164784e216555d930210371279ba8e").traverse_commits())[0]
    assert len(commit.modifications) == 1
    commit = list(RepositoryMining('test-repos/whitespace',
                                   skip_whitespaces=True,
                                   single="338a74ceae164784e216555d930210371279ba8e").traverse_commits())[0]
    assert len(commit.modifications) == 0


@pytest.mark.parametrize('path', ["test-repos/whitespace"])
def test_ignore_add_whitespaces_and_modified_normal_line(git_repo):
    commit = list(RepositoryMining('test-repos/whitespace',
                                   single="52716ef1f11e07308b5df1b313aec5496d5e91ce").traverse_commits())[0]
    assert len(commit.modifications) == 1
    parsed_normal_diff = commit.modifications[0].diff_parsed
    commit = list(RepositoryMining('test-repos/whitespace',
                                   skip_whitespaces=True,
                                   single="52716ef1f11e07308b5df1b313aec5496d5e91ce").traverse_commits())[0]
    assert len(commit.modifications) == 1
    parsed_wo_whitespaces_diff = commit.modifications[0].diff_parsed
    assert len(parsed_normal_diff['added']) == 2
    assert len(parsed_wo_whitespaces_diff['added']) == 1

    assert len(parsed_normal_diff['deleted']) == 1
    assert len(parsed_wo_whitespaces_diff['deleted']) == 0


def test_ignore_deleted_whitespaces():
    commit = list(RepositoryMining('test-repos/whitespace',
                                   single="e6e429f6b485e18fb856019d9953370fd5420b20").traverse_commits())[0]
    assert len(commit.modifications) == 1
    commit = list(RepositoryMining('test-repos/whitespace',
                                   skip_whitespaces=True,
                                   single="e6e429f6b485e18fb856019d9953370fd5420b20").traverse_commits())[0]
    assert len(commit.modifications) == 0


def test_ignore_add_whitespaces_and_changed_file():
    commit = list(RepositoryMining('test-repos/whitespace',
                                   single="532068e9d64b8a86e07eea93de3a57bf9e5b4ae0").traverse_commits())[0]
    assert len(commit.modifications) == 2
    commit = list(RepositoryMining('test-repos/whitespace',
                                   skip_whitespaces=True,
                                   single="532068e9d64b8a86e07eea93de3a57bf9e5b4ae0").traverse_commits())[0]
    assert len(commit.modifications) == 1


def test_clone_repo_to(tmp_path):
    dt2 = datetime(2018, 10, 20)
    url = "https://github.com/ishepard/pydriller.git"
    assert len(list(RepositoryMining(
        path_to_repo=url,
        to=dt2,
        clone_repo_to=str(tmp_path)).traverse_commits())) == 159


def test_clone_repo_to_not_existing():
    with pytest.raises(Exception):
        list(RepositoryMining("https://github.com/ishepard/pydriller",
                              clone_repo_to="NOTEXISTINGDIR").traverse_commits())



from pathlib import Path
from datetime import datetime

import pytest

from pydriller.metrics.process.contributors_count import ContributorsCount

TEST_DATA = [
   ('test-repos/pydriller', 'pydriller/git_repository.py', '8b69cae085581256adfdbd58c0e499395819b84d', '115953109b57d841ccd0952d70f8ed6703d175cd', 2),
   ('test-repos/pydriller', 'domain/modification.py', '71e053f61fc5d31b3e31eccd9c79df27c31279bf', 'ab36bf45859a210b0eae14e17683f31d19eea041', 1)
]

@pytest.mark.parametrize('path_to_repo, filepath, from_commit, to_commit, expected', TEST_DATA)
def test_with_commits(path_to_repo, filepath, from_commit, to_commit, expected):
    metric = ContributorsCount(path_to_repo=path_to_repo,
                               from_commit=from_commit,
                               to_commit=to_commit)

    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected

TEST_DATA = [
   ('test-repos/pydriller', 'pydriller/git_repository.py', datetime(2019, 12, 17), datetime(2019, 12, 24), 2),
   ('test-repos/pydriller', 'domain/modification.py', datetime(2018, 3, 21), datetime(2018, 3, 27), 1)
]

@pytest.mark.parametrize('path_to_repo, filepath, since, to, expected', TEST_DATA)
def test_with_dates(path_to_repo, filepath, since, to, expected):
    metric = ContributorsCount(path_to_repo=path_to_repo,
                               since=since,
                               to=to)

    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected
